# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Kmodify
-------

The ``drgn.helpers.experimental.kmodify`` module provides experimental helpers
for modifying the state of the running kernel. This works by loading a
temporary kernel module, so the kernel must support loadable kernel modules
(``CONFIG_MODULES=y``) and allow loading unsigned modules
(``CONFIG_MODULE_SIG_FORCE=n``). It is currently only implemented for x86-64.

.. warning::
    These helpers are powerful but **extremely** dangerous. Use them with care.
"""

import ctypes
import errno
import operator
import os
import random
import re
import string
import struct
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

if TYPE_CHECKING:
    from _typeshed import SupportsWrite

    if sys.version_info < (3, 11):
        from typing_extensions import assert_never
    else:
        from typing import assert_never  # novermin

from _drgn_util.elf import ET, SHF, SHN, SHT, STB, STT, STV
from _drgn_util.platform import SYS
from drgn import (
    Architecture,
    FaultError,
    IntegerLike,
    Object,
    ObjectAbsentError,
    PlatformFlags,
    PrimitiveType,
    Program,
    ProgramFlags,
    Type,
    TypeKind,
    alignof,
    cast,
    implicit_convert,
    offsetof,
    sizeof,
)
from drgn.helpers.common.prog import takes_program_or_default

__all__ = (
    "call_function",
    "call_functions",
    "pass_pointer",
    "write_memory",
    "write_object",
)


_c = ctypes.CDLL(None, use_errno=True)
_syscall = _c.syscall
_syscall.restype = ctypes.c_long


# os.memfd_create() was added in Python 3.8, and only with glibc >= 2.27
# (manylinux2014 uses glibc 2.17). The syscall was added in Linux 3.17, so fall
# back to using it directly.
if hasattr(os, "memfd_create"):
    _memfd_create = os.memfd_create  # novermin
else:

    def _memfd_create(
        name: str,
        flags: int = 1,  # MFD_CLOEXEC
    ) -> int:
        fd = _syscall(
            ctypes.c_long(SYS["memfd_create"]),
            ctypes.c_char_p(os.fsencode(name)),
            ctypes.c_uint(flags),
        )
        if fd < 0:
            errnum = ctypes.get_errno()
            raise OSError(errnum, os.strerror(errnum))
        return fd


class _ElfSection:
    def __init__(
        self,
        *,
        name: str,
        type: SHT,
        flags: SHF = SHF(0),
        data: bytes,
        addr: int = 0,
        link: int = 0,
        info: int = 0,
        addralign: int = 1,
        entsize: int = 0,
    ) -> None:
        self.name = name
        self.type = type
        self.flags = flags
        self.data = data
        self.addr = addr
        self.link = link
        self.info = info
        self.addralign = addralign
        self.entsize = entsize


class _ElfSymbol(NamedTuple):
    name: str
    value: int
    size: int
    section: Union[str, SHN]
    type: STT
    binding: STB
    visibility: STV = STV.DEFAULT


class _ElfRelocation(NamedTuple):
    offset: int
    type: int
    symbol_name: str
    section_symbol: bool
    addend: int = 0


def _write_elf(
    file: "SupportsWrite[bytes]",
    *,
    machine: int,
    is_little_endian: bool,
    is_64_bit: bool,
    rela: bool,
    sections: Sequence[_ElfSection],
    symbols: Sequence[_ElfSymbol],
    relocations: Mapping[str, Sequence[_ElfRelocation]],
) -> None:
    endian = "<" if is_little_endian else ">"
    if is_64_bit:
        ehdr_struct = struct.Struct(endian + "16BHHIQQQIHHHHHH")
        shdr_struct = struct.Struct(endian + "IIQQQQIIQQ")
        rela_struct = struct.Struct(endian + "QQq")

        def r_info(sym: int, type: int) -> int:
            return (sym << 32) | type

        sym_struct = struct.Struct(endian + "IBBHQQ")

        def sym_fields(sym: _ElfSymbol) -> Tuple[int, int, int, int, int]:
            return (
                (sym.binding << 4) + (sym.type & 0xF),
                sym.visibility,
                (
                    section_name_to_index[sym.section]
                    if isinstance(sym.section, str)
                    else sym.section
                ),
                sym.value,
                sym.size,
            )

    else:
        ehdr_struct = struct.Struct(endian + "16BHHIIIIIHHHHHH")
        shdr_struct = struct.Struct(endian + "10I")
        rela_struct = struct.Struct(endian + "IIi")

        def r_info(sym: int, type: int) -> int:
            return (sym << 8) | type

        sym_struct = struct.Struct(endian + "IIIBBH")

        def sym_fields(sym: _ElfSymbol) -> Tuple[int, int, int, int, int]:
            return (
                sym.value,
                sym.size,
                (sym.binding << 4) + (sym.type & 0xF),
                sym.visibility,
                (
                    section_name_to_index[sym.section]
                    if isinstance(sym.section, str)
                    else sym.section
                ),
            )

    section_symbols = [
        _ElfSymbol(
            name="",
            value=0,
            size=0,
            type=STT.SECTION,
            binding=STB.LOCAL,
            section=section.name,
        )
        for section in sections
        if section.type == SHT.PROGBITS
    ]
    section_name_to_symbol_index = {
        sym.section: i for i, sym in enumerate(section_symbols, 1)
    }
    symbol_name_to_index = {
        sym.name: i for i, sym in enumerate(symbols, 1 + len(section_symbols))
    }
    section_symbols.extend(symbols)
    symbols = section_symbols
    del section_symbols

    def relocation_symbol_index(reloc: _ElfRelocation) -> int:
        if reloc.section_symbol:
            return section_name_to_symbol_index[reloc.symbol_name]
        else:
            return symbol_name_to_index[reloc.symbol_name]

    if rela:
        reloc_prefix = ".rela"
        reloc_sht = SHT.RELA
        reloc_size = rela_struct.size

        def relocation_data(relocations: Sequence[_ElfRelocation]) -> bytes:
            data = bytearray(len(relocations) * rela_struct.size)
            for i, relocation in enumerate(relocations):
                rela_struct.pack_into(
                    data,
                    i * rela_struct.size,
                    relocation.offset,
                    r_info(
                        relocation_symbol_index(relocation),
                        relocation.type,
                    ),
                    relocation.addend,
                )
            return data

    else:
        raise NotImplementedError("SHT_REL relocations")

    symtab_section_index = 1 + len(sections) + len(relocations)

    sections = list(sections)
    i = 0
    while i < len(sections):
        section = sections[i]
        try:
            section_relocations = relocations[section.name]
        except KeyError:
            i += 1
            continue
        sections.insert(
            i + 1,
            _ElfSection(
                name=reloc_prefix + section.name,
                type=reloc_sht,
                flags=SHF.INFO_LINK,
                data=relocation_data(section_relocations),
                link=symtab_section_index,
                info=i + 1,
                addralign=8 if is_64_bit else 4,
                entsize=reloc_size,
            ),
        )
        i += 2

    section_name_to_index = {section.name: i for i, section in enumerate(sections, 1)}

    if len(sections) < symtab_section_index - 1:
        raise ValueError(
            f"relocations for unknown section {', '.join(relocations.keys() - section_name_to_index)}"
        )

    symtab_data = bytearray((len(symbols) + 1) * sym_struct.size)
    strtab_data = bytearray(1)

    sym_local_end = 1
    for i, sym in enumerate(symbols, 1):
        if sym.name:
            st_name = len(strtab_data)
            strtab_data.extend(sym.name.encode())
            strtab_data.append(0)
        else:
            st_name = 0

        sym_struct.pack_into(
            symtab_data, i * sym_struct.size, st_name, *sym_fields(sym)
        )
        if sym.binding == STB.LOCAL:
            if sym_local_end != i:
                raise ValueError("local symbol after non-local symbol")
            sym_local_end = i + 1

    sections.append(
        _ElfSection(
            name=".symtab",
            type=SHT.SYMTAB,
            data=symtab_data,
            link=len(sections) + 2,
            info=sym_local_end,
            entsize=sym_struct.size,
        )
    )
    sections.append(_ElfSection(name=".strtab", type=SHT.STRTAB, data=strtab_data))

    shstrtab_data = bytearray(1)
    sh_name = []
    for section in sections:
        sh_name.append(len(shstrtab_data))
        shstrtab_data.extend(section.name.encode())
        shstrtab_data.append(0)
    sh_name.append(len(shstrtab_data))
    shstrtab_data.extend(b".shstrtab\0")
    sections.append(_ElfSection(name=".shstrtab", type=SHT.STRTAB, data=shstrtab_data))

    shnum = len(sections) + 1  # + 1 for the SHT_NULL section
    headers_size = ehdr_struct.size + shdr_struct.size * shnum
    file.write(
        ehdr_struct.pack(
            0x7F,  # ELFMAG0
            ord("E"),  # ELFMAG1
            ord("L"),  # ELFMAG2
            ord("F"),  # ELFMAG3
            2 if is_64_bit else 1,  # EI_CLASS = ELFCLASS64 or ELFCLASS32
            1 if is_little_endian else 2,  # EI_DATA = ELFDATA2LSB or ELFDATA2MSB
            1,  # EI_VERSION = EV_CURRENT
            0,  # EI_OSABI = ELFOSABI_NONE
            0,  # EI_ABIVERSION
            0,
            0,
            0,
            0,
            0,
            0,
            0,  # EI_PAD
            ET.REL,  # e_type
            machine,
            1,  # e_version = EV_CURRENT
            0,  # e_entry
            0,  # e_phoff
            ehdr_struct.size,  # e_shoff
            0,  # e_flags
            ehdr_struct.size,  # e_ehsize
            0,  # e_phentsize
            0,  # e_phnum
            shdr_struct.size,  # e_shentsize
            shnum,  # e_shnum
            shnum - 1,  # e_shstrndx
        )
    )

    # SHT_NULL section.
    file.write(bytes(shdr_struct.size))

    section_data_offset = headers_size
    for i, section in enumerate(sections):
        section_data_offset += -section_data_offset % section.addralign
        file.write(
            shdr_struct.pack(
                sh_name[i],  # sh_name
                section.type,  # sh_type
                section.flags,  # sh_flags
                section.addr,  # sh_addr
                section_data_offset,  # sh_offset
                len(section.data),  # sh_size
                section.link,  # sh_link
                section.info,  # sh_info
                section.addralign,  # sh_addralign
                section.entsize,  # sh_entsize
            )
        )
        section_data_offset += len(section.data)

    section_data_offset = headers_size
    for section in sections:
        padding = -section_data_offset % section.addralign
        if padding:
            file.write(bytes(padding))
        section_data_offset += padding
        file.write(section.data)
        section_data_offset += len(section.data)


# Abstract syntax tree-ish representation of code to inject.
class _Integer:
    def __init__(self, size: int, value: IntegerLike) -> None:
        self.size = size
        self.value = operator.index(value)


class _Symbol(NamedTuple):
    name: str
    offset: int = 0
    section: bool = False


class _Call(NamedTuple):
    func: _Symbol
    args: Sequence[Union[_Integer, _Symbol]]


class _StoreReturnValue(NamedTuple):
    size: int
    dst: _Symbol


class _Return(NamedTuple):
    value: _Integer


class _ReturnIfLastReturnValueNonZero(NamedTuple):
    value: _Integer


class _AtomicSetBit(NamedTuple):
    nr: int
    address: _Integer


class _AtomicClearBit(NamedTuple):
    nr: int
    address: _Integer


_FunctionBodyNode = Union[
    _Call,
    _StoreReturnValue,
    _Return,
    _ReturnIfLastReturnValueNonZero,
    _AtomicSetBit,
    _AtomicClearBit,
]


class _Function(NamedTuple):
    body: Sequence[_FunctionBodyNode]


class _CodeGen_x86_64:
    _R_X86_64_PC32 = 2
    _R_X86_64_PLT32 = 4
    _R_X86_64_32S = 11

    _rax = 0
    _r11 = 11

    _argument_registers = (
        7,  # rdi
        6,  # rsi,
        2,  # rdx
        1,  # rcx
        8,  # r8
        9,  # r9
    )

    def __init__(self) -> None:
        self.code = bytearray()
        self.relocations: List[_ElfRelocation] = []
        self._epilogue_jumps: List[int] = []

    def enter_frame(self, size: int) -> None:
        if size < 0:
            raise ValueError("invalid stack frame size")

        self.code.extend(
            # endbr64
            # This is only needed if CONFIG_X86_KERNEL_IBT=y, but it's much
            # simpler to do it unconditionally, and it's a no-op if not needed.
            b"\xF3\x0F\x1E\xFA"
            # Set up the frame pointer.
            # push %rbp
            b"\x55"
            # mov %rsp, %rbp
            b"\x48\x89\xE5"
        )

        # The System V ABI requires that rsp % 16 == 0 on function entry. We
        # need to make sure that rsp % 16 == 8 in the function body so that the
        # return address pushed by the call will make rsp % 16 == 0. push %rbp
        # makes rsp % 16 == 8. So, we need to align the requested size up to 16
        # bytes.
        size = (size + 15) & ~15
        if size > 0:
            # sub $size, %rsp
            if size < 128:
                self.code.extend(b"\x48\x83\xEC")
                self.code.append(size)
            else:
                self.code.extend(b"\x48\x81\xEC")
                self.code.extend(size.to_bytes(4, "little", signed=True))

    def leave_frame(self) -> None:
        # Fix up all of the jumps to the epilogue.
        for offset in self._epilogue_jumps:
            self.code[offset - 4 : offset] = (len(self.code) - offset).to_bytes(
                4, "little", signed=True
            )

        self.code.extend(
            # leave
            b"\xC9"
            # ret
            b"\xC3"
        )

    def _mov_imm(self, i: _Integer, reg: int, sign_extend_bits: int = 0) -> None:
        value = i.value & ((1 << max(i.size * 8, sign_extend_bits)) - 1)
        assert value >= 0 and value <= 0xFFFFFFFFFFFFFFFF
        assert reg < 16
        if value <= 0xFFFFFFFF:
            if reg >= 8:
                self.code.append(0x41)  # REX.B
                reg -= 8
            self.code.append(0xB8 + reg)
            self.code.extend(value.to_bytes(4, "little"))
        else:
            rex = 0x48  # REX.W
            if reg >= 8:
                rex |= 1  # REX.B
                reg -= 8
            self.code.append(rex)
            if value >= 0xFFFFFFFF80000000:
                self.code.append(0xC7)
                self.code.append(0xC0 + reg)
                self.code.extend((value & 0xFFFFFFFF).to_bytes(4, "little"))
            else:
                self.code.append(0xB8 + reg)
                self.code.extend(value.to_bytes(8, "little"))

    def _mov_symbol(self, sym: _Symbol, reg: int) -> None:
        rex = 0x48  # REX.W
        if reg >= 8:
            rex |= 1  # REX.B
            reg -= 8
        self.code.append(rex)
        self.code.append(0xC7)
        self.code.append(0xC0 + reg)
        self.relocations.append(
            _ElfRelocation(
                offset=len(self.code),
                type=self._R_X86_64_32S,
                symbol_name=sym.name,
                section_symbol=sym.section,
                addend=sym.offset,
            )
        )
        self.code.extend(bytes(4))

    def _store_rax_on_stack(self, offset: int) -> None:
        # mov %rax, offset(%rsp)
        if offset == 0:
            self.code.extend(b"\x48\x89\x04\x24")
        elif -128 <= offset < 128:
            self.code.extend(b"\x48\x89\x44\x24")
            self.code.append(offset & 0xFF)
        else:
            self.code.extend(b"\x48\x89\x84\x24")
            self.code.extend(offset.to_bytes(4, "little", signed=True))

    def _store_imm_on_stack(self, i: _Integer, offset: int) -> None:
        value = i.value & ((1 << max(i.size * 8, 64)) - 1)
        if (0 <= value <= 0x7FFFFFFF) or (
            0xFFFFFFFF80000000 <= value <= 0xFFFFFFFFFFFFFFFF
        ):
            # mov $value, offset(%rsp)
            if offset == 0:
                self.code.extend(b"\x48\xC7\x04\x24")
            elif -128 <= offset < 128:
                self.code.extend(b"\x48\xC7\x44\x24")
                self.code.append(offset & 0xFF)
            else:
                self.code.extend(b"\x48\xC7\x84\x24")
                self.code.extend(offset.to_bytes(4, "little", signed=True))
            self.code.extend((value & 0xFFFFFFFF).to_bytes(4, "little"))
        else:
            self._mov_imm(i, self._rax, 64)
            self._store_rax_on_stack(offset)

    def _store_symbol_on_stack(self, sym: _Symbol, offset: int) -> None:
        self._mov_symbol(sym, self._rax)
        self._store_rax_on_stack(offset)

    def call(self, func: _Symbol, args: Sequence[Union[_Integer, _Symbol]]) -> None:
        for i, arg in enumerate(args):
            if i < len(self._argument_registers):
                reg = self._argument_registers[i]
                if isinstance(arg, _Integer):
                    # Clang/LLVM as of version 19 relies on <32-bit arguments
                    # being sign-extended to 32 bits despite this not being
                    # guaranteed by the psABI. It's unclear whether this will
                    # be resolved by changing LLVM or the psABI, so work around
                    # it for now. See:
                    # https://groups.google.com/g/x86-64-abi/c/h7FFh30oS3s/m/Gksanh3WAAAJ
                    # https://github.com/llvm/llvm-project/issues/12579
                    # https://gcc.gnu.org/bugzilla/show_bug.cgi?id=46942
                    self._mov_imm(arg, reg, 32)
                else:
                    self._mov_symbol(arg, reg)
            else:
                stack_offset = 8 * (i - len(self._argument_registers))
                if isinstance(arg, _Integer):
                    self._store_imm_on_stack(arg, stack_offset)
                else:
                    self._store_symbol_on_stack(arg, stack_offset)

        # call near
        self.code.append(0xE8)
        self.relocations.append(
            _ElfRelocation(
                offset=len(self.code),
                type=self._R_X86_64_PLT32,
                symbol_name=func.name,
                section_symbol=func.section,
                addend=-4,
            )
        )
        self.code.extend(bytes(4))

    def store_return_value(self, size: int, dst: _Symbol) -> None:
        if size == 1:
            # movb %al, ...
            self.code.extend(b"\x88\x05")
        elif size == 2:
            # movw %ax, ...
            self.code.extend(b"\x66\x89\x05")
        elif size == 4:
            # movl %eax, ...
            self.code.extend(b"\x89\x05")
        elif size == 8:
            # movq %rax, ...
            self.code.extend(b"\x48\x89\x05")
        else:
            raise NotImplementedError("{size}-byte return values not implemented")
        self.relocations.append(
            _ElfRelocation(
                offset=len(self.code),
                type=self._R_X86_64_PC32,
                symbol_name=dst.name,
                section_symbol=dst.section,
                addend=dst.offset - 4,
            )
        )
        # ... 0x0(%rip)
        self.code.extend(bytes(4))

    def return_(self, value: _Integer, last: bool) -> None:
        if value.size > 8:
            raise NotImplementedError(
                "return values larger than 8 bytes not implemented"
            )
        self._mov_imm(value, self._rax)
        # Jump to the function epilogue. If this return is the last operation,
        # we can fall through instead of jumping.
        if not last:
            # jmp
            self.code.extend(b"\xE9\x00\x00\x00\x00")
            # The destination needs to be fixed up later.
            self._epilogue_jumps.append(len(self.code))

    def return_if_last_return_value_nonzero(self, value: _Integer) -> None:
        if value.size > 8:
            raise NotImplementedError(
                "return values larger than 8 bytes not implemented"
            )
        # mov %rax, %rdx
        self.code.extend(b"\x48\x89\xC2")
        self._mov_imm(value, self._rax)
        # Jump to the function epilogue if the last return value was non-zero.
        self.code.extend(
            # test %rdx, %rdx
            b"\x48\x85\xD2"
            # jnz
            b"\x0F\x85\x00\x00\x00\x00"
        )
        # The destination needs to be fixed up later.
        self._epilogue_jumps.append(len(self.code))

    def atomic_set_bit(self, nr: int, address: _Integer) -> None:
        # mov $address, %rax
        self._mov_imm(address, self._rax)
        # lock; orb $(1 << nr), (%rax)
        self.code.extend(b"\xf0\x80\x08")
        self.code.append(1 << nr)

    def atomic_clear_bit(self, nr: int, address: _Integer) -> None:
        # mov $address, %rax
        self._mov_imm(address, self._rax)
        # lock; andb $~(1 << nr), (%rax)
        self.code.extend(b"\xf0\x80\x20")
        self.code.append(0xFF ^ (1 << nr))


class _Arch_X86_64:
    ELF_MACHINE = 62  # EM_X86_64
    RELA = True
    ABSOLUTE_ADDRESS_RELOCATION_TYPE = 1  # R_X86_64_64

    @staticmethod
    def code_gen(func: _Function) -> Tuple[bytes, Sequence[_ElfRelocation]]:
        needed_stack_size = 0
        for node in func.body:
            if not isinstance(node, _Call):
                continue
            stack_size = len(_CodeGen_x86_64._argument_registers) * -8
            for arg in node.args:
                if isinstance(arg, _Integer):
                    if arg.size > 8:
                        raise NotImplementedError(
                            "passing integers larger than 8 bytes not implemented"
                        )
                    stack_size += 8
                elif isinstance(arg, _Symbol):
                    stack_size += 8
                else:
                    assert_never(arg)
            if stack_size > needed_stack_size:
                needed_stack_size = stack_size

        code_gen = _CodeGen_x86_64()

        code_gen.enter_frame(needed_stack_size)

        for i, node in enumerate(func.body):
            if isinstance(node, _Call):
                code_gen.call(node.func, node.args)
            elif isinstance(node, _StoreReturnValue):
                code_gen.store_return_value(node.size, node.dst)
            elif isinstance(node, _Return):
                code_gen.return_(node.value, last=i == len(func.body) - 1)
            elif isinstance(node, _ReturnIfLastReturnValueNonZero):
                code_gen.return_if_last_return_value_nonzero(node.value)
            elif isinstance(node, _AtomicSetBit):
                code_gen.atomic_set_bit(node.nr, node.address)
            elif isinstance(node, _AtomicClearBit):
                code_gen.atomic_clear_bit(node.nr, node.address)
            else:
                assert_never(node)

        code_gen.leave_frame()

        return code_gen.code, code_gen.relocations


def _find_exported_symbol_in_section(
    prog: Program, name: bytes, start: int, stop: int
) -> int:
    kernel_symbol_type = prog.type("struct kernel_symbol")
    if kernel_symbol_type.has_member("name_offset"):

        def kernel_symbol_name(sym: Object) -> Object:
            return cast("char *", sym.name_offset.address_of_()) + sym.name_offset

    else:

        def kernel_symbol_name(sym: Object) -> Object:
            return sym.name

    syms = Object(prog, prog.pointer_type(kernel_symbol_type), start)
    lo = 0
    hi = (stop - start) // sizeof(kernel_symbol_type)
    while lo < hi:
        mid = (lo + hi) // 2
        sym_name = kernel_symbol_name(syms[mid]).string_()
        if sym_name < name:
            lo = mid + 1
        elif sym_name > name:
            hi = mid
        else:
            return mid
    return -1


# If CONFIG_MODVERSIONS=y, then we need a __versions section containing a CRC
# of each exported symbol that we use. Since we intentionally don't use any
# symbols, we only need it for the special module_layout symbol.
def _get_versions_section(struct_module: Type) -> Optional[_ElfSection]:
    prog = struct_module.prog
    try:
        return prog.cache["kmodify___versions_section"]
    except KeyError:
        pass

    # module_layout is defined if and only if CONFIG_MODVERSIONS=y.
    have_module_layout = False
    try:
        have_module_layout = prog["module_layout"].address_ is not None
    except KeyError:
        pass

    if have_module_layout:
        # We only check the non-GPL-only section because module_layout is
        # non-GPL-only.
        i = _find_exported_symbol_in_section(
            prog,
            b"module_layout",
            prog.symbol("__start___ksymtab").address,
            prog.symbol("__stop___ksymtab").address,
        )
        if i < 0:
            raise LookupError("module_layout not found")

        # Since Linux kernel commit 71810db27c1c ("modversions: treat symbol
        # CRCs as 32 bit quantities") (in v4.10), CRCs are in an array of s32.
        # Before that, they are in an array of unsigned long. Determine the
        # correct type from struct module::crcs.
        module_layout_crc = (
            Object(
                prog,
                struct_module.member("crcs").type,
                prog.symbol("__start___kcrctab").address,
            )[i].value_()
            & 0xFFFFFFFF
        )

        struct_modversion_info = prog.type("struct modversion_info")
        section = _ElfSection(
            name="__versions",
            type=SHT.PROGBITS,
            flags=SHF.ALLOC,
            data=Object(
                prog,
                struct_modversion_info,
                {
                    "crc": module_layout_crc,
                    "name": b"module_layout",
                },
            ).to_bytes_(),
            addralign=alignof(struct_modversion_info),
        )
    else:
        section = None
    prog.cache["kmodify___versions_section"] = section
    return section


class _Kmodify:
    def __init__(self, prog: Program) -> None:
        if prog.flags & (
            ProgramFlags.IS_LINUX_KERNEL | ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL
        ) != (
            ProgramFlags.IS_LINUX_KERNEL | ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL
        ):
            raise ValueError("kmodify is only available for the running kernel")
        platform = prog.platform
        if platform is None:
            raise ValueError("program platform is not known")
        self.prog = prog
        self.is_little_endian = bool(platform.flags & PlatformFlags.IS_LITTLE_ENDIAN)
        self.is_64_bit = bool(platform.flags & PlatformFlags.IS_64_BIT)

        if platform.arch == Architecture.X86_64:
            # When we add support for another architecture, we're going to need
            # an _Arch Protocol.
            self.arch = _Arch_X86_64
        else:
            raise NotImplementedError(
                f"kmodify not implemented for {platform.arch.name} architecture"
            )

    _KMOD_NAME_CHARS = string.digits + string.ascii_letters

    def insert(
        self,
        *,
        name: str,
        code: bytes,
        code_relocations: Sequence[_ElfRelocation],
        data: bytes,
        data_alignment: int,
        symbols: Sequence[_ElfSymbol],
    ) -> int:
        struct_module = self.prog.type("struct module")

        module_name = "".join(
            [
                "drgn_kmodify_",
                # Randomize to avoid name collisions.
                *random.choices(self._KMOD_NAME_CHARS, k=12),
                "_",
                name,
            ]
        ).encode("ascii")[: sizeof(struct_module.member("name").type) - 1]

        sections = [
            _ElfSection(
                name=".init.text",
                type=SHT.PROGBITS,
                flags=SHF.ALLOC | SHF.EXECINSTR,
                data=code,
                # This should be good enough for any supported architecture.
                addralign=16,
            ),
            _ElfSection(
                name=".data",
                type=SHT.PROGBITS,
                flags=SHF.WRITE | SHF.ALLOC,
                data=data,
                addralign=data_alignment,
            ),
            _ElfSection(
                name=".gnu.linkonce.this_module",
                type=SHT.PROGBITS,
                flags=SHF.WRITE | SHF.ALLOC,
                data=Object(
                    self.prog, struct_module, {"name": module_name}
                ).to_bytes_(),
                addralign=alignof(struct_module),
            ),
            _ElfSection(
                name=".modinfo",
                type=SHT.PROGBITS,
                flags=SHF.ALLOC,
                data=b"".join(
                    [
                        b"%b=%b\0" % (key, value)
                        for key, value in (
                            (b"license", b"GPL"),
                            (b"depends", b""),
                            # A retpoline kernel complains when loading a
                            # non-retpoline module. We never make indirect
                            # calls, so we can claim to be a retpoline module.
                            # (Note that it's harmless to set this for
                            # non-retpoline kernels.)
                            (b"retpoline", b"Y"),
                            (b"name", module_name),
                            (b"vermagic", self.prog["vermagic"].string_()),
                        )
                    ]
                ),
            ),
            # This section and its associated symbols are required if memory
            # allocation profiling is enabled and is ignored otherwise.
            _ElfSection(
                name=".codetag.alloc_tags",
                type=SHT.PROGBITS,
                flags=SHF.WRITE | SHF.ALLOC,
                data=b"",
            ),
        ]

        # Add the __versions section if needed.
        versions_section = _get_versions_section(struct_module)
        if versions_section is not None:
            sections.append(versions_section)

        symbols = [
            *symbols,
            _ElfSymbol(
                name="init_module",
                value=0,
                size=len(code),
                type=STT.FUNC,
                binding=STB.GLOBAL,
                section=".init.text",
            ),
            _ElfSymbol(
                name="__start_alloc_tags",
                value=0,
                size=0,
                type=STT.NOTYPE,
                binding=STB.GLOBAL,
                section=".codetag.alloc_tags",
            ),
            _ElfSymbol(
                name="__stop_alloc_tags",
                value=0,
                size=0,
                type=STT.NOTYPE,
                binding=STB.GLOBAL,
                section=".codetag.alloc_tags",
            ),
        ]

        relocations = {
            ".init.text": code_relocations,
            ".gnu.linkonce.this_module": [
                _ElfRelocation(
                    offset=offsetof(struct_module, "init"),
                    type=self.arch.ABSOLUTE_ADDRESS_RELOCATION_TYPE,
                    symbol_name="init_module",
                    section_symbol=False,
                )
            ],
        }

        with open(_memfd_create(module_name.decode() + ".ko"), "wb") as f:
            _write_elf(
                f,
                machine=self.arch.ELF_MACHINE,
                is_little_endian=self.is_little_endian,
                is_64_bit=self.is_64_bit,
                rela=self.arch.RELA,
                sections=sections,
                symbols=symbols,
                relocations=relocations,
            )
            f.flush()

            if _syscall(
                ctypes.c_long(SYS["finit_module"]),
                ctypes.c_int(f.fileno()),
                ctypes.c_char_p(b""),
                ctypes.c_int(0),
            ):
                return -ctypes.get_errno()
            else:
                return 0


_COPY_TO_FROM_KERNEL_NOFAULT_NAMES = (
    # Names used since Linux kernel commit fe557319aa06 ("maccess: rename
    # probe_kernel_{read,write} to copy_{from,to}_kernel_nofault") (in
    # v5.8-rc2).
    ("copy_to_kernel_nofault", "copy_from_kernel_nofault"),
    # Names used before Linux kernel commit 48c49c0e5f31 ("maccess: remove
    # various unused weak aliases") (in v5.8-rc1).
    ("__probe_kernel_write", "__probe_kernel_read"),
    # Names briefly used between those two commits.
    ("probe_kernel_write", "probe_kernel_read"),
)


@takes_program_or_default
def write_memory(prog: Program, address: IntegerLike, value: bytes) -> None:
    """
    Write a byte string to kernel memory.

    >>> os.uname().sysname
    'Linux'
    >>> write_memory(prog["init_uts_ns"].name.sysname.address_, b"Lol\\0")
    >>> os.uname().sysname
    'Lol'

    .. warning::
        This attempts to detect writes to bad addresses and raise a
        :class:`~drgn.FaultError`, but this is best-effort and may still crash
        the kernel. Writing bad data can of course also cause a crash when the
        data is used. Additionally, this is not atomic, so the data may be
        accessed while it is partially written.

    :param address: Address to write to.
    :param value: Byte string to write.
    :raises FaultError: if the address cannot be written to
    """
    copy_to_kernel_nofault_address = None
    copy_from_kernel_nofault_address = None
    for (
        copy_to_kernel_nofault,
        copy_from_kernel_nofault,
    ) in _COPY_TO_FROM_KERNEL_NOFAULT_NAMES:
        try:
            copy_to_kernel_nofault_address = prog[copy_to_kernel_nofault].address_
            copy_from_kernel_nofault_address = prog[copy_from_kernel_nofault].address_
            break
        except KeyError:
            pass
    if copy_to_kernel_nofault_address is None:
        raise LookupError("copy_to_kernel_nofault not found")
    if copy_from_kernel_nofault_address is None:
        raise LookupError("copy_from_kernel_nofault not found")

    kmodify = _Kmodify(prog)
    address = operator.index(address)
    sizeof_int = sizeof(prog.type("int"))
    sizeof_void_p = sizeof(prog.type("void *"))
    sizeof_size_t = sizeof(prog.type("size_t"))
    code, code_relocations = kmodify.arch.code_gen(
        _Function(
            [
                # copy_to_kernel_nofault() can still fault in some cases; see
                # https://lore.kernel.org/all/f0e171cbae576758d9387cee374dd65088e75b07.1725223574.git.osandov@fb.com/
                # copy_from_kernel_nofault() catches some of those cases.
                _Call(
                    _Symbol(copy_from_kernel_nofault),
                    [
                        _Symbol(".data", section=True, offset=len(value)),
                        _Integer(sizeof_void_p, address),
                        _Integer(sizeof_size_t, 1),
                    ],
                ),
                _ReturnIfLastReturnValueNonZero(
                    _Integer(sizeof_int, -errno.EFAULT),
                ),
                _Call(
                    _Symbol(copy_to_kernel_nofault),
                    [
                        _Integer(sizeof_void_p, address),
                        _Symbol(".data", section=True),
                        _Integer(sizeof_size_t, len(value)),
                    ],
                ),
                _ReturnIfLastReturnValueNonZero(
                    _Integer(sizeof_int, -errno.EFAULT),
                ),
                _Return(_Integer(sizeof_int, -errno.EINPROGRESS)),
            ]
        )
    )
    ret = kmodify.insert(
        name=f"write_{len(value)}",
        code=code,
        code_relocations=code_relocations,
        data=value + b"\0",
        # Align generously so that the copy can use larger units and small
        # copies can be slightly less racy.
        data_alignment=16,
        symbols=[
            _ElfSymbol(
                name=copy_to_kernel_nofault,
                value=copy_to_kernel_nofault_address,
                size=0,
                type=STT.FUNC,
                binding=STB.LOCAL,
                section=SHN.ABS,
            ),
            _ElfSymbol(
                name=copy_from_kernel_nofault,
                value=copy_from_kernel_nofault_address,
                size=0,
                type=STT.FUNC,
                binding=STB.LOCAL,
                section=SHN.ABS,
            ),
        ],
    )
    if ret != -errno.EINPROGRESS:
        if ret == -errno.EFAULT:
            raise FaultError("could not write to memory", address)
        elif ret:
            raise OSError(-ret, os.strerror(-ret))
        else:
            raise ValueError("module init did not run")


def _modify_bit(prog: Program, nr: int, address: int, value: bool) -> None:
    # This and the codegen representations are all in terms of bytes.
    assert 0 <= nr < 8

    copy_from_kernel_nofault_address = None
    for _, copy_from_kernel_nofault in _COPY_TO_FROM_KERNEL_NOFAULT_NAMES:
        try:
            copy_from_kernel_nofault_address = prog[copy_from_kernel_nofault].address_
            break
        except KeyError:
            pass
    if copy_from_kernel_nofault_address is None:
        raise LookupError("copy_from_kernel_nofault not found")
    sizeof_int = sizeof(prog.type("int"))
    sizeof_void_p = sizeof(prog.type("void *"))
    sizeof_size_t = sizeof(prog.type("size_t"))

    kmodify = _Kmodify(prog)
    if not kmodify.is_little_endian:
        # Big-endian needs different calculations for {set,clear}_bit(), and
        # I'm not sure about bit fields. kmodify only supports little-endian
        # architectures at the moment anyways.
        raise NotImplementedError("_modify_bit() is only implemented for little-endian")
    code, code_relocations = kmodify.arch.code_gen(
        _Function(
            [
                _Call(
                    # Catch invalid addresses instead of crashing. Note that
                    # this won't catch read-only addresses, unfortunately.
                    _Symbol(copy_from_kernel_nofault),
                    [
                        _Symbol(".data", section=True),
                        _Integer(sizeof_void_p, address),
                        _Integer(sizeof_size_t, 1),
                    ],
                ),
                _ReturnIfLastReturnValueNonZero(
                    _Integer(sizeof_int, -errno.EFAULT),
                ),
                (_AtomicSetBit if value else _AtomicClearBit)(
                    nr, _Integer(sizeof_void_p, address)
                ),
                _Return(_Integer(sizeof_int, -errno.EINPROGRESS)),
            ]
        )
    )
    ret = kmodify.insert(
        name="set_bit" if value else "clear_bit",
        code=code,
        code_relocations=code_relocations,
        data=b"\0",
        data_alignment=1,
        symbols=[
            _ElfSymbol(
                name=copy_from_kernel_nofault,
                value=copy_from_kernel_nofault_address,
                size=0,
                type=STT.FUNC,
                binding=STB.LOCAL,
                section=SHN.ABS,
            ),
        ],
    )
    if ret != -errno.EINPROGRESS:
        if ret == -errno.EFAULT:
            raise FaultError("could not write to memory", address)
        elif ret:
            raise OSError(-ret, os.strerror(-ret))
        else:
            raise ValueError("module init did not run")


def _set_or_clear_bit(nr: IntegerLike, bitmap: Object, value: bool) -> None:
    nr = operator.index(nr)
    address = implicit_convert("unsigned long *", bitmap).value_() + (nr // 8)
    nr %= 8
    _modify_bit(bitmap.prog_, nr, address, value)


def set_bit(nr: IntegerLike, bitmap: Object) -> None:
    """
    Atomically set a bit in kernel memory.

    .. warning::
        This attempts to detect bad addresses and raise a
        :class:`~drgn.FaultError`, but this is best-effort and may still crash
        the kernel. In particular, this function will crash the kernel if the
        target address is read-only. Writing an invalid bit can of course also
        cause a crash when the data is used.

    :param nr: Bit number.
    :param bitmap: ``unsigned long *``
    """
    _set_or_clear_bit(nr, bitmap, True)


def clear_bit(nr: IntegerLike, bitmap: Object) -> None:
    """
    Atomically clear a bit in kernel memory.

    .. warning::
        The warnings about :func:`set_bit()` also apply to ``clear_bit()``.

    :param nr: Bit number.
    :param bitmap: ``unsigned long *``
    """
    _set_or_clear_bit(nr, bitmap, False)


def write_object(
    object: Object, value: Any, *, dereference: Optional[bool] = None
) -> None:
    """
    Write to an object in kernel memory.

    >>> os.system("uptime -p")
    up 12 minutes
    >>> write_object(prog["init_time_ns"].offsets.boottime.tv_sec, 1000000000)
    >>> os.system("uptime -p")
    up 3 decades, 1 year, 37 weeks, 1 hour, 59 minutes

    Bit fields are currently only supported if they are either byte-aligned or
    a single bit. Writes to a bit field are atomic with respect to other bit
    fields.

    .. warning::
        The warnings about :func:`write_memory()` and :func:`set_bit()` also
        apply to ``write_object()``.

    :param object: Object to write to.
    :param value: Value to write. This may be an :class:`~drgn.Object` or a
        Python value. Either way, it will be converted to the type of *object*.
    :param dereference: If *object* is a pointer, whether to dereference it. If
        ``True``, then write to the object pointed to by *object*
        (``*ptr = value``). If ``False``, then write to the pointer itself
        (``ptr = value``). This is a common source of confusion, so it is
        required if *object* is a pointer.
    :raises ValueError: if *object* is not a reference object (i.e., its
        address is not known)
    :raises TypeError: if *object* is a pointer and *dereference* is not given
    :raises TypeError: if *object* is not a pointer and *dereference* is
        ``True``
    :raises NotImplementedError: if *object* is a bit field that is not
        byte-aligned or a single bit
    """
    type = object.type_
    if type.unaliased_kind() == TypeKind.POINTER:
        if dereference is None:
            raise TypeError(
                "to write to pointed-to object (*ptr = value), use dereference=True; "
                "to write to pointer itself (ptr = value), use dereference=False"
            )
        elif dereference:
            object = object[0]
    elif dereference:
        raise TypeError("object is not a pointer")

    address = object.address_
    if address is None:
        raise ValueError("cannot write to value object")

    bit_field_size = object.bit_field_size_
    if isinstance(value, Object):
        value = implicit_convert(type, value, bit_field_size=bit_field_size)
    else:
        value = Object(object.prog_, type, value, bit_field_size=bit_field_size)

    if bit_field_size is None:
        bit_field_size = 0
    bit_offset: int = object.bit_offset_  # type: ignore[assignment]  # address_ is not None, so bit_offset_ is not, either.

    if bit_field_size % 8 == 0 and bit_offset % 8 == 0:
        write_memory(object.prog_, address, value.to_bytes_())
    elif bit_field_size == 1:
        _modify_bit(object.prog_, bit_offset, address, bool(value))
    else:
        raise NotImplementedError("only byte-aligned or 1-bit bit fields are supported")


def _default_argument_promotions(obj: Object) -> Object:
    type = obj.type_.unaliased()
    if type.kind == TypeKind.INT:
        return +obj
    elif type.primitive == PrimitiveType.C_FLOAT:
        return cast("double", obj)
    else:
        return obj


def _resolve_function(prog: Program, func: Union[str, Object]) -> Tuple[Type, Object]:
    if not isinstance(func, Object):
        func = prog.function(func)

    func_type = func.type_.unaliased()
    try:
        if func_type.kind == TypeKind.FUNCTION:
            func_pointer = func.address_of_()
        elif func_type.kind == TypeKind.POINTER:
            func_type = func_type.type.unaliased()
            if func_type.kind != TypeKind.FUNCTION:
                raise TypeError("func must be function or function pointer")
            func_pointer = func.read_()
        else:
            raise TypeError("func must be function or function pointer")
    except ObjectAbsentError:
        raise ObjectAbsentError("function is absent, likely inlined") from None

    return func_type, func_pointer


def _prepare_call_args(
    prog: Program,
    func_type: Type,
    func_pointer: Object,
    args: Sequence[Any],
    data: bytearray,
    data_alignment: int,
    out_pointers: List[Tuple["pass_pointer", int]],
) -> Tuple[List[Union[_Integer, _Symbol]], int]:
    return_type = func_type.type.unaliased()
    if return_type.kind not in {
        TypeKind.VOID,
        TypeKind.INT,
        TypeKind.BOOL,
        TypeKind.ENUM,
        TypeKind.POINTER,
    }:
        raise NotImplementedError(f"{return_type} return values not implemented")

    if len(args) < len(func_type.parameters):
        raise TypeError(f"not enough arguments for {func_pointer}; got {len(args)}")
    if not func_type.is_variadic and len(args) > len(func_type.parameters):
        raise TypeError(f"too many arguments for {func_pointer}; got {len(args)}")

    call_args: List[Union[_Integer, _Symbol]] = []

    def align_data(alignment: int) -> None:
        nonlocal data_alignment
        if alignment > data_alignment:
            data_alignment = alignment
        data.extend(bytes(-len(data) % alignment))

    for i, arg in enumerate(args):
        if i < len(func_type.parameters):
            parameter_type = func_type.parameters[i].type.unaliased()

        if (
            (
                isinstance(arg, (str, bytes, bytearray))
                or (
                    isinstance(arg, pass_pointer)
                    and isinstance(arg.object, (str, bytes, bytearray))
                )
            )
            and i < len(func_type.parameters)
            and parameter_type.kind == TypeKind.POINTER
            and parameter_type.type.unaliased().primitive
            in (
                PrimitiveType.C_CHAR,
                PrimitiveType.C_SIGNED_CHAR,
                PrimitiveType.C_UNSIGNED_CHAR,
            )
        ):
            # Convert strings to null-terminated character arrays.
            if not isinstance(arg, pass_pointer):
                arg = pass_pointer(arg)
            if isinstance(arg.object, str):
                arg.object = arg.object.encode()
            arg.object = Object(
                prog,
                prog.array_type(
                    parameter_type.type,
                    len(arg.object) + 1,
                    language=func_type.language,
                ),
                arg.object,
            )
        elif isinstance(arg, Object) and arg.type_.unaliased_kind() == TypeKind.ARRAY:
            # Convert arrays to pointers.
            if arg.address_ is None:
                arg = pass_pointer(arg)
            else:
                arg = arg + 0

        if isinstance(arg, pass_pointer):
            if not isinstance(arg.object, Object):
                arg.object = Object(prog, value=arg.object)
            type = arg.object.type_
            unaliased_type = type.unaliased()
            if unaliased_type.kind == TypeKind.ARRAY:
                type = unaliased_type.type
            if i < len(func_type.parameters):
                # We don't need the result, just type checking.
                implicit_convert(
                    func_type.parameters[i].type,
                    Object(prog, prog.pointer_type(type), 0),
                )
            value = arg.object.to_bytes_()

            align_data(alignof(arg.object.type_))
            out_pointers.append((arg, len(data)))
            call_args.append(_Symbol(".data", section=True, offset=len(data)))
            data.extend(value)
        else:
            if not isinstance(arg, Object):
                arg = Object(prog, value=arg)
            if i < len(func_type.parameters):
                arg = implicit_convert(func_type.parameters[i].type, arg)
            else:
                arg = _default_argument_promotions(arg)

            type = arg.type_.unaliased()
            if type.kind not in {
                TypeKind.INT,
                TypeKind.BOOL,
                TypeKind.ENUM,
                TypeKind.POINTER,
            }:
                if type.kind in {
                    TypeKind.FLOAT,
                    TypeKind.STRUCT,
                    TypeKind.UNION,
                    TypeKind.CLASS,
                }:
                    raise NotImplementedError(
                        f"passing {type} by value not implemented"
                    )
                else:
                    raise ValueError(f"cannot pass {type} by value")

            call_args.append(_Integer(sizeof(type), arg.value_()))

    return call_args, data_alignment


def _insert_call_function(
    kmodify: _Kmodify,
    *,
    name: str,
    function_body: List[_FunctionBodyNode],
    symbols: List[_ElfSymbol],
    data: bytearray,
    data_alignment: int,
    out_pointers: List[Tuple["pass_pointer", int]],
) -> Any:
    prog = kmodify.prog

    # copy_to_user() is the more obvious choice, but it's an inline function.
    copy_to_user_nofault_address = None
    for copy_to_user_nofault in (
        # Name used since Linux kernel commit c0ee37e85e0e ("maccess: rename
        # probe_user_{read,write} to copy_{from,to}_user_nofault") (in
        # v5.8-rc2).
        "copy_to_user_nofault",
        # Name used before Linux kernel commit 48c49c0e5f31 ("maccess: remove
        # various unused weak aliases") (in v5.8-rc1).
        "__probe_user_write",
        # Name briefly used between those two commits.
        "probe_user_write",
    ):
        try:
            copy_to_user_nofault_address = prog[copy_to_user_nofault].address_
            break
        except KeyError:
            continue
    if copy_to_user_nofault_address is None:
        raise LookupError("copy_to_user_nofault not found")

    sizeof_int = sizeof(prog.type("int"))
    out_buf = None
    if data:
        out_buf = ctypes.create_string_buffer(len(data))
        function_body.append(
            _Call(
                _Symbol(copy_to_user_nofault),
                [
                    _Integer(
                        sizeof(prog.type("void *")),
                        ctypes.addressof(out_buf),
                    ),
                    _Symbol(".data", section=True),
                    _Integer(sizeof(prog.type("size_t")), len(data)),
                ],
            )
        )
        function_body.append(
            _ReturnIfLastReturnValueNonZero(_Integer(sizeof_int, -errno.EFAULT))
        )
        symbols.append(
            _ElfSymbol(
                name=copy_to_user_nofault,
                value=copy_to_user_nofault_address,
                size=0,
                type=STT.FUNC,
                binding=STB.LOCAL,
                section=SHN.ABS,
            )
        )

    function_body.append(_Return(_Integer(sizeof_int, -errno.EINPROGRESS)))

    code, code_relocations = kmodify.arch.code_gen(_Function(function_body))

    ret = kmodify.insert(
        name=name,
        code=code,
        code_relocations=code_relocations,
        data=data,
        data_alignment=data_alignment,
        symbols=symbols,
    )
    if ret != -errno.EINPROGRESS:
        if ret:
            raise OSError(-ret, os.strerror(-ret))
        else:
            raise ValueError("module init did not run")

    for out_pointer, offset in out_pointers:
        out_pointer.object = Object.from_bytes_(
            prog,
            out_pointer.object.type_,
            out_buf,  # type: ignore[arg-type]  # out_buf can't be None here.
            bit_offset=offset * 8,
        )

    return out_buf


@takes_program_or_default
def call_function(prog: Program, func: Union[str, Object], *args: Any) -> Object:
    """
    Call a function in the kernel.

    >>> task = find_task(99)
    >>> if task:
    ...     call_function("wake_up_process", task)
    ...
    (int)1

    Arguments can be either :class:`~drgn.Object`\\ s or Python values. The
    function return value is returned as an :class:`~drgn.Object`:

    >>> # GFP_KERNEL isn't in the kernel debug info
    >>> # We have to use this trick to get it.
    >>> for flag in prog["gfpflag_names"]:
    ...     if flag.name.string_() == b"GFP_KERNEL":
    ...             GFP_KERNEL = flag.mask
    ...             break
    ...
    >>> # kmalloc() is actually a macro.
    >>> # We have to call the underlying function.
    >>> p = call_function("__kmalloc_noprof", 13, GFP_KERNEL)
    >>> p
    (void *)0xffff991701ef43c0
    >>> identify_address(p)
    'slab object: kmalloc-16+0x0'
    >>> call_function("kfree", p)
    (void)<absent>
    >>> identify_address(p)
    'free slab object: kmalloc-16+0x0'

    Variadic functions are also supported:

    >>> call_function("_printk", "Hello, world! %d\\n", Object(prog, "int", 1234))
    (int)18
    >>> os.system("dmesg | tail -1")
    [ 1138.223004] Hello, world! 1234

    Constructed values can be passed by pointer using :class:`pass_pointer()`:

    >>> sb = prog["init_fs"].root.mnt.mnt_sb
    >>> sb.s_shrink.scan_objects
    (unsigned long (*)(struct shrinker *, struct shrink_control *))super_cache_scan+0x0 = 0xffffffffbda4c487
    >>> sc = pass_pointer(Object(prog, "struct shrink_control",
    ...                   {"gfp_mask": GFP_KERNEL, "nr_to_scan": 100, "nr_scanned": 100}))
    >>> call_function(sb.s_shrink.scan_objects, sb.s_shrink, sc)
    (unsigned long)31

    If the function modifies the passed value, the :class:`pass_pointer` object
    is updated:

    >>> sc.object
    (struct shrink_control){
            .gfp_mask = (gfp_t)3264,
            .nid = (int)0,
            .nr_to_scan = (unsigned long)1,
            .nr_scanned = (unsigned long)100,
            .memcg = (struct mem_cgroup *)0x0,
    }

    .. note::
        It is not possible to call some functions, including inlined functions
        and function-like macros. If the unavailable function is a wrapper
        around another function, sometimes the wrapped function can be called
        instead.

    .. warning::
        Calling a function incorrectly may cause the kernel to crash or
        misbehave in various ways.

        The function is called from process context. Note that the function may
        have context, locking, or reference counting requirements.

    :param func: Function to call. May be a function name, function object, or
        function pointer object.
    :param args: Function arguments. :class:`int`, :class:`float`, and
        :class:`bool` arguments are converted as "literals" with
        ``Object(prog, value=...)``. :class:`str` and :class:`bytes` arguments
        are converted to ``char`` array objects. :class:`pass_pointer`
        arguments are copied to the kernel, passed by pointer, and copied back.
    :return: Function return value.
    :raises TypeError: if the passed arguments have incorrect types for the
        function
    :raises ObjectAbsentError: if the function cannot be called because it is
        inlined
    :raises LookupError: if a function with the given name is not found
        (possibly because it is actually a function-like macro)
    """
    kmodify = _Kmodify(prog)
    func_type, func_pointer = _resolve_function(prog, func)

    out_pointers: List[Tuple[pass_pointer, int]] = []
    data = bytearray()
    call_args, data_alignment = _prepare_call_args(
        prog, func_type, func_pointer, args, data, 1, out_pointers
    )

    function_body: List[_FunctionBodyNode] = [_Call(_Symbol("func"), call_args)]
    symbols = [
        _ElfSymbol(
            name="func",
            value=func_pointer.value_(),
            size=0,
            type=STT.FUNC,
            binding=STB.LOCAL,
            section=SHN.ABS,
        )
    ]

    return_type = func_type.type.unaliased()
    if return_type.kind != TypeKind.VOID:
        alignment = alignof(return_type)
        if alignment > data_alignment:
            data_alignment = alignment
        data.extend(bytes(-len(data) % alignment))
        return_offset = len(data)
        return_size = sizeof(return_type)
        data.extend(bytes(return_size))
        function_body.append(
            _StoreReturnValue(
                return_size,
                _Symbol(".data", section=True, offset=return_offset),
            )
        )

    kmod_name = "call"
    try:
        symbol_name_match = re.match(r"[0-9a-zA-Z_]+", prog.symbol(func_pointer).name)
        if symbol_name_match:
            kmod_name = "call_" + symbol_name_match.group()
    except LookupError:
        pass

    out_buf = _insert_call_function(
        kmodify,
        name=kmod_name,
        function_body=function_body,
        symbols=symbols,
        data=data,
        data_alignment=data_alignment,
        out_pointers=out_pointers,
    )

    if return_type.kind == TypeKind.VOID:
        return Object(prog, func_type.type)
    else:
        return Object.from_bytes_(
            prog, func_type.type, out_buf, bit_offset=return_offset * 8
        )


@takes_program_or_default
def call_functions(prog: Program, calls: Sequence[Sequence[Any]]) -> None:
    """
    Call multiple functions in the kernel.

    This is more efficient than calling :func:`call_function()` multiple times
    because it only inserts one kernel module.

    >>> leaked_namespaces
    [Object(prog, 'struct net *', value=0xffff8f568869af80), ...]
    >>> call_functions([("netns_put", ns) for ns in leaked_namespaces])

    Function return values are ignored. Use :func:`call_function()` if you need
    the return value.

    .. warning::
        See :func:`call_function()` for warnings about calling kernel
        functions.

    :param calls: List of calls, where each call is a tuple of
        ``(func, *args)``. See :func:`call_function()` for details on *func*
        and *args*.
    """
    kmodify = _Kmodify(prog)

    function_body: List[_FunctionBodyNode] = []
    symbols: List[_ElfSymbol] = []
    out_pointers: List[Tuple[pass_pointer, int]] = []
    data = bytearray()
    data_alignment = 1

    first_func_value = None
    multiple_func_values = False
    for call_i, call in enumerate(calls):
        func_type, func_pointer = _resolve_function(prog, call[0])

        call_args, data_alignment = _prepare_call_args(
            prog, func_type, func_pointer, call[1:], data, data_alignment, out_pointers
        )

        symbol_name = f"func{call_i}"
        function_body.append(_Call(_Symbol(symbol_name), call_args))
        func_value = func_pointer.value_()
        symbols.append(
            _ElfSymbol(
                name=symbol_name,
                value=func_value,
                size=0,
                type=STT.FUNC,
                binding=STB.LOCAL,
                section=SHN.ABS,
            )
        )

        if first_func_value is None:
            first_func_value = func_value
        elif func_value != first_func_value:
            multiple_func_values = True

    kmod_name = "call"
    if first_func_value is not None:
        try:
            symbol_name_match = re.match(
                r"[0-9a-zA-Z_]+", prog.symbol(first_func_value).name
            )
            if symbol_name_match:
                kmod_name = "call_" + symbol_name_match.group()
        except LookupError:
            pass

        if multiple_func_values:
            if kmod_name == "call":
                kmod_name = "call_multiple"
            else:
                kmod_name += "_and_more"
        elif len(calls) > 1:
            kmod_name += f"_{len(calls)}x"

    _insert_call_function(
        kmodify,
        name=kmod_name,
        function_body=function_body,
        symbols=symbols,
        data=data,
        data_alignment=data_alignment,
        out_pointers=out_pointers,
    )


class pass_pointer:
    object: Any
    """
    Wrapped object. Updated to an :class:`~drgn.Object` containing the final
    value after the function call.
    """

    def __init__(self, object: Any) -> None:
        """
        Wrapper used to pass values to :func:`call_function()` or
        :func:`call_functions()` by pointer.

        :param object: :class:`~drgn.Object` or Python value to wrap.
        """
        self.object = object

    def __repr__(self) -> str:
        return f"pass_pointer({self.object!r})"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Commands for getting system information."""

import argparse
import collections
import datetime
import functools
import itertools
import logging
import sys
from typing import (
    Any,
    Callable,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from drgn import Architecture, Object, Program
from drgn.commands import (
    _repr_black,
    argument,
    argument_group,
    drgn_argument,
    mutually_exclusive_group,
)
from drgn.commands._crash.common import (
    Cpuspec,
    CrashDrgnCodeBuilder,
    _crash_get_panic_context,
    _format_seconds_duration,
    crash_command,
    crash_get_context,
    parse_cpuspec,
)
from drgn.helpers.common.format import (
    CellFormat,
    RowOptions,
    double_quote_ascii_string,
    escape_ascii_string,
    number_in_binary_units,
    print_table,
)
from drgn.helpers.linux.block import (
    disk_name,
    for_each_disk,
    request_queue_busy_iter,
    rq_data_dir,
)
from drgn.helpers.linux.cpumask import (
    cpumask_to_cpulist,
    num_online_cpus,
    num_present_cpus,
)
from drgn.helpers.linux.device import (
    MAJOR,
    for_each_registered_blkdev,
    for_each_registered_chrdev,
)
from drgn.helpers.linux.ioport import for_each_resource
from drgn.helpers.linux.irq import (
    for_each_irq_desc,
    gate_desc_func,
    irq_desc_action_names,
    irq_desc_affinity_mask,
    irq_desc_chip_name,
    irq_to_desc,
)
from drgn.helpers.linux.kconfig import _get_raw_kconfig
from drgn.helpers.linux.mm import totalram_pages
from drgn.helpers.linux.panic import panic_message, panic_task, tainted
from drgn.helpers.linux.pci import (
    PCI_EXP_TYPE,
    for_each_pci_root_bus,
    pci_bus_for_each_child,
    pci_bus_for_each_dev,
    pci_bus_name,
    pci_is_bridge,
    pci_name,
    pci_pcie_type,
)
from drgn.helpers.linux.percpu import per_cpu
from drgn.helpers.linux.pid import for_each_task
from drgn.helpers.linux.printk import print_dmesg
from drgn.helpers.linux.sched import (
    get_task_state,
    loadavg,
    task_cpu,
    task_on_cpu,
    task_thread_info,
)
from drgn.helpers.linux.timekeeping import (
    ktime_get_boottime_seconds,
    ktime_get_real_seconds,
)

logger = logging.getLogger("drgn")


def _append_kernel(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.panic", "tainted")
    code.append(
        """\
try:
    kernel = prog.main_module().debug_file_path
except LookupError:
    kernel = None

taints = tainted()
"""
    )


def _get_kernel(prog: Program) -> Optional[str]:
    try:
        kernel = prog.main_module().debug_file_path
    except LookupError:
        kernel = None
    if tainted(prog):
        if kernel is None:
            kernel = "[TAINTED]"
        else:
            kernel += "  [TAINTED]"
    return kernel


def _append_dumpfile(code: CrashDrgnCodeBuilder) -> None:
    code.append("dumpfile = prog.core_dump_path\n")


def _get_dumpfile(prog: Program) -> Optional[str]:
    return prog.core_dump_path


def _append_cpus(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import(
        "drgn.helpers.linux.cpumask", "num_online_cpus", "num_present_cpus"
    )
    code.append(
        """\
cpus = num_present_cpus()
offline_cpus = cpus - num_online_cpus()
"""
    )


def _get_cpus(prog: Program) -> str:
    present = num_present_cpus(prog)
    online = num_online_cpus(prog)
    cpus = str(present)
    if present > online:
        cpus += f" [OFFLINE: {present - online}]"
    return cpus


def _append_date(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.timekeeping", "ktime_get_real_seconds")
    code.append("timestamp = ktime_get_real_seconds().value_()\n")


def _get_date(prog: Program) -> str:
    timestamp = ktime_get_real_seconds(prog).value_()
    dt = datetime.datetime.fromtimestamp(timestamp).astimezone()
    return dt.strftime("%a %b %e %T %Z %Y")


def _append_uptime(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.timekeeping", "uptime")
    code.append("uptime_ = uptime()\n")


def _get_uptime(prog: Program) -> str:
    return _format_seconds_duration(ktime_get_boottime_seconds(prog).value_())


def _append_load_average(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.sched", "loadavg")
    code.append("load_average = loadavg()\n")


def _get_load_average(prog: Program) -> str:
    return ", ".join([f"{v:.2f}" for v in loadavg(prog)])


def _append_tasks(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.pid", "for_each_task")
    code.append("num_tasks = sum(1 for _ in for_each_task(idle=True))\n")


def _get_tasks(prog: Program) -> str:
    return str(sum(1 for _ in for_each_task(prog, idle=True)))


def _append_utsname_field(code: CrashDrgnCodeBuilder, field: str) -> None:
    code.append(f'{field} = prog["init_uts_ns"].name.{field}.string_()\n')


def _get_utsname_field(prog: Program, field: str) -> str:
    return escape_ascii_string(
        getattr(prog["init_uts_ns"].name, field).string_(),
        escape_backslash=True,
    )


def _append_release(code: CrashDrgnCodeBuilder) -> None:
    code.append('release = prog["UTS_RELEASE"].string_()\n')


def _get_release(prog: Program) -> str:
    return escape_ascii_string(prog["UTS_RELEASE"].string_(), escape_backslash=True)


def _append_memory(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.mm", "totalram_pages")
    code.append('memory = totalram_pages() * prog["PAGE_SIZE"].value_()\n')


def _get_memory(prog: Program) -> str:
    return number_in_binary_units(totalram_pages(prog) * prog["PAGE_SIZE"])


def _append_panic(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.panic", "panic_message")
    code.append("panic = panic_message()\n")


def _get_panic(prog: Program) -> Optional[str]:
    message = panic_message(prog)
    if message is None:
        return message
    return double_quote_ascii_string(message)


def _append_pid(code: CrashDrgnCodeBuilder) -> None:
    code.append("pid = task.pid\n")


def _get_pid(task: Object) -> str:
    return str(task.pid.value_())


def _append_command(code: CrashDrgnCodeBuilder) -> None:
    code.append("comm = task.comm\n")


def _get_command(task: Object) -> str:
    return double_quote_ascii_string(task.comm.string_())


def _append_task(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.sched", "task_thread_info")
    code.append("thread_info = task_thread_info(task)\n")


def _get_task(task: Object) -> str:
    return f"{task.value_():x}  [THREAD_INFO: {task_thread_info(task).value_():x}]"


def _append_cpu(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.sched", "task_cpu")
    code.append("cpu = task_cpu(task)\n")


def _get_cpu(task: Object) -> str:
    return str(task_cpu(task))


def _append_state(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.panic", "panic_task")
    code.add_from_import("drgn.helpers.linux.sched", "get_task_state", "task_on_cpu")
    code.append(
        """\
state = get_task_state(task)
is_active = task_on_cpu(task)
is_panic = task == panic_task()
"""
    )


def _get_state(task: Object) -> str:
    state = get_task_state(task)
    if task == panic_task(task.prog_):
        state += " (PANIC)"
    elif task_on_cpu(task):
        state += " (ACTIVE)"
    return state


_FIELDS: List[
    Tuple[str, Callable[[CrashDrgnCodeBuilder], None], Callable[[Program], Any]]
] = [
    ("KERNEL", _append_kernel, _get_kernel),
    ("DUMPFILE", _append_dumpfile, _get_dumpfile),
    ("CPUS", _append_cpus, _get_cpus),
    ("DATE", _append_date, _get_date),
    ("UPTIME", _append_uptime, _get_uptime),
    ("LOAD AVERAGE", _append_load_average, _get_load_average),
    ("TASKS", _append_tasks, _get_tasks),
    (
        "NODENAME",
        functools.partial(_append_utsname_field, field="nodename"),
        functools.partial(_get_utsname_field, field="nodename"),
    ),
    ("RELEASE", _append_release, _get_release),
    (
        "VERSION",
        functools.partial(_append_utsname_field, field="version"),
        functools.partial(_get_utsname_field, field="version"),
    ),
    (
        "MACHINE",
        functools.partial(_append_utsname_field, field="machine"),
        functools.partial(_get_utsname_field, field="machine"),
    ),
    ("MEMORY", _append_memory, _get_memory),
    ("PANIC", _append_panic, _get_panic),
]

_TASK_FIELDS: List[
    Tuple[str, Callable[[CrashDrgnCodeBuilder], None], Callable[[Object], Any]]
] = [
    ("PID", _append_pid, _get_pid),
    ("COMMAND", _append_command, _get_command),
    ("TASK", _append_task, _get_task),
    ("CPU", _append_cpu, _get_cpu),
    ("STATE", _append_state, _get_state),
]


def _append_sys_drgn_system(code: CrashDrgnCodeBuilder) -> None:
    for _, append, _ in _FIELDS:
        append(code)


def _append_sys_drgn_task(code: CrashDrgnCodeBuilder) -> None:
    for _, append, _ in _TASK_FIELDS:
        append(code)


def _print_sys(
    prog: Program,
    system_fields: bool = True,
    context: Union[Literal["panic", "current"], Object, None] = None,
) -> Optional[Object]:
    if isinstance(context, Object) or context is None:
        task = context
    elif context == "panic":
        try:
            task = _crash_get_panic_context(prog)
        except LookupError:
            logger.warning("panic task not found")
            task = None
    elif context == "current":
        task = crash_get_context(prog)
    else:
        raise ValueError("invalid context")

    rows = []

    if system_fields:
        for name, _, getter in _FIELDS:
            try:
                value = getter(prog)
            except Exception as e:
                logger.warning("couldn't get %s: %s", name, e)
                continue
            if value is not None:
                rows.append((CellFormat(name, ">"), value))

    if task is not None:
        for name, _, task_getter in _TASK_FIELDS:
            try:
                value = task_getter(task)
            except Exception as e:
                logger.warning("couldn't get %s: %s", name, e)
                continue
            if value is not None:
                rows.append((CellFormat(name, ">"), value))

    print_table(rows, sep=": ")
    return task


@crash_command(
    description="system information",
    arguments=(
        argument(
            "config",
            metavar="config",
            choices=("config",),
            nargs="?",
            help="print kernel configuration (requires ``CONFIG_IKCONFIG``)",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_sys(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.config:
        if args.drgn:
            sys.stdout.write(
                """\
from drgn.helpers.linux.kconfig import get_kconfig


kconfig = get_kconfig()
"""
            )
            return
        sys.stdout.write(_get_raw_kconfig(prog).decode())
        return

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        _append_sys_drgn_system(code)
        return code.print()
    _print_sys(prog)


def _print_ioports(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        sys.stdout.write(
            """\
from drgn.helpers.linux.ioport import for_each_resource


for resource in for_each_resource(prog["ioport_resource"].address_of_()):
    start = resource.start
    end = resource.end
    name = resource.name
"""
        )
        return

    rows: List[Sequence[Any]] = [
        (
            CellFormat("RESOURCE", "^"),
            CellFormat("RANGE", "^"),
            "NAME",
        )
    ]
    ioport_resource = prog["ioport_resource"]
    width = len(f"{ioport_resource.end.value_():x}")
    for resource in for_each_resource(ioport_resource.address_of_()):
        rows.append(
            (
                CellFormat(resource.value_(), "^x"),
                f"{resource.start.value_():0{width}x}-{resource.end.value_():0{width}x}",
                escape_ascii_string(resource.name.string_(), escape_backslash=True),
            )
        )
    print_table(rows)


def _append_pci_devs(rows: List[Sequence[Any]], bus: Object) -> None:
    first = True
    self = bus.self.read_()
    for dev in itertools.chain(((self,) if self else ()), pci_bus_for_each_dev(bus)):
        if first:
            rows.append(
                RowOptions(
                    (
                        "",
                        CellFormat("PCI DEV", "^"),
                        "DO:BU:SL.FN",
                        CellFormat("CLASS", ">"),
                        CellFormat("PCI_ID", "^"),
                        "TYPE",
                    ),
                    group=1,
                )
            )
            first = False

        type = pci_pcie_type(dev)
        type_str = type.name if isinstance(type, PCI_EXP_TYPE) else f"{type:x}"
        if pci_is_bridge(dev):
            type_str += " [BRIDGE]"
        rows.append(
            RowOptions(
                (
                    "",
                    CellFormat(dev.value_(), "^x"),
                    pci_name(dev),
                    CellFormat(f"{dev.member_('class').value_() >> 8:04x}", ">"),
                    f"{dev.vendor.value_():04x}:{dev.device.value_():04x}",
                    type_str,
                ),
                group=1,
            )
        )


def _append_pci_buses(
    rows: List[Sequence[Any]], parent_bus: Object, want_blank_line: bool
) -> None:
    for bus in pci_bus_for_each_child(parent_bus):
        if want_blank_line:
            rows.append(())
        want_blank_line = True

        rows.append(
            RowOptions(
                (
                    "",
                    CellFormat("PCI BUS", "^"),
                    CellFormat("PARENT BUS", "^"),
                ),
                group=2,
            )
        )
        rows.append(
            RowOptions(
                (
                    "",
                    CellFormat(bus.value_(), "^x"),
                    CellFormat(parent_bus.value_(), "^x"),
                ),
                group=2,
            )
        )
        _append_pci_devs(rows, bus)
        _append_pci_buses(rows, bus, True)


def _print_pci(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        sys.stdout.write(
            """\
from drgn.helpers.linux.pci import (
    for_each_pci_root_bus,
    pci_bus_for_each_child,
    pci_bus_for_each_dev,
    pci_bus_name,
    pci_is_bridge,
    pci_name,
    pci_pcie_type,
)


def visit_pci_dev(dev):
    do_bu_sl_fn = pci_name(dev)
    class_ = dev.member_("class") >> 8
    pci_id = f"{dev.vendor.value_():04x}:{dev.device.value_():04x}"
    type_ = pci_pcie_type(dev)
    is_bridge = pci_is_bridge(dev)


def visit_pci_bus(bus):
    for dev in pci_bus_for_each_dev(bus):
        visit_pci_dev(dev)
    for child_bus in pci_bus_for_each_child(bus):
        visit_pci_bus(child_bus)


for root_bus in for_each_pci_root_bus():
    busname = pci_bus_name(root_bus)
    visit_pci_bus(root_bus)
"""
        )
        return

    rows: List[Sequence[Any]] = []
    first = True
    for bus in for_each_pci_root_bus(prog):
        if first:
            first = False
        else:
            rows.append(())

        rows.append(
            (
                CellFormat("ROOT BUS", "^"),
                "BUSNAME",
            )
        )
        rows.append(
            (
                CellFormat(bus.value_(), "^x"),
                pci_bus_name(bus),
            )
        )
        _append_pci_devs(rows, bus)
        _append_pci_buses(rows, bus, False)
    print_table(rows)


def _print_disks(prog: Program, drgn_arg: bool, only_busy: bool) -> None:
    if drgn_arg:
        sys.stdout.write(
            """\
from drgn.helpers.linux.block import (
    disk_name,
    for_each_disk,
    request_queue_busy_iter,
    rq_data_dir,
)


for disk in for_each_disk():
    major = disk.major
    name = disk_name(disk)
    queue = disk.queue
    read = write = 0
    for rq in request_queue_busy_iter(queue):
        if rq_data_dir(rq):
            write += 1
        else:
            read += 1
    total = read + write
"""
        )
        return

    rows: List[Sequence[Any]] = [
        (
            CellFormat("MAJOR", ">"),
            "GENDISK",
            "NAME",
            "REQUEST_QUEUE",
            CellFormat("TOTAL", ">"),
            CellFormat("READ", ">"),
            CellFormat("WRITE", ">"),
        )
    ]
    for disk in for_each_disk(prog):
        queue = disk.queue.read_()
        num_ops = [0, 0]
        for rq in request_queue_busy_iter(queue):
            num_ops[rq_data_dir(rq)] += 1
        total_ops = sum(num_ops)
        if not total_ops and only_busy:
            continue

        rows.append(
            (
                disk.major.value_(),
                CellFormat(disk.value_(), "<x"),
                escape_ascii_string(disk_name(disk), escape_backslash=True),
                CellFormat(disk.queue.value_(), "<x"),
                total_ops,
                *num_ops,
            ),
        )
    print_table(rows)


@crash_command(
    description="devices",
    arguments=(
        argument_group(
            mutually_exclusive_group(
                argument(
                    "-i",
                    dest="ioport",
                    action="store_true",
                    help="display I/O port regions",
                ),
                argument(
                    "-p",
                    dest="pci",
                    action="store_true",
                    help="display PCI devices",
                ),
                argument(
                    "-d",
                    dest="disk",
                    action="store_true",
                    help="display all disks and their number of in-progress requests",
                ),
                argument(
                    "-D",
                    dest="busy_disk",
                    action="store_true",
                    help="""
                    like **-d**, but only display disks that have in-progress
                    requests
                    """,
                ),
            ),
            title="type",
            description="""
            What type of devices to display. If not given, display registered
            character and block device majors.
            """,
        ),
        drgn_argument,
    ),
)
def _crash_cmd_dev(
    prog: Program, command_name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.ioport:
        return _print_ioports(prog, args.drgn)
    elif args.pci:
        return _print_pci(prog, args.drgn)
    elif args.disk or args.busy_disk:
        return _print_disks(prog, args.drgn, args.busy_disk)

    if args.drgn:
        sys.stdout.write(
            """\
from drgn.helpers.linux.block import for_each_disk
from drgn.helpers.linux.device import (
    MAJOR,
    for_each_registered_blkdev,
    for_each_registered_chrdev,
)


for dev, _, name, cdev in for_each_registered_chrdev():
    major = MAJOR(dev)
    if cdev:
        operations = cdev.ops


for major, name, _ in for_each_registered_blkdev():
    pass


for disk in for_each_disk():
    major = disk.major
    operations = disk.fops
"""
        )
        return

    rows: List[Sequence[Any]] = [
        (
            "CHRDEV",
            "NAME",
            CellFormat("CDEV", "^"),
            "OPERATIONS",
        )
    ]
    for dev, _, name, cdev in for_each_registered_chrdev(prog):
        operations_cell: Any = ""
        if cdev:
            cdev_cell = CellFormat(cdev.value_(), "^x")
            ops = cdev.ops.value_()
            try:
                operations_cell = prog.symbol(ops).name
            except LookupError:
                operations_cell = CellFormat(ops, "x")
        else:
            cdev_cell = CellFormat("(none)", "^")
        rows.append(
            (
                MAJOR(dev),
                escape_ascii_string(name, escape_backslash=True),
                cdev_cell,
                operations_cell,
            )
        )

    rows.append(())

    major_to_gendisk = collections.defaultdict(list)
    for disk in for_each_disk(prog):
        major_to_gendisk[disk.major.value_()].append(disk)
    major_to_gendisk.default_factory = None

    rows.append(
        (
            "BLKDEV",
            "NAME",
            CellFormat("GENDISK", "^"),
            "OPERATIONS",
        )
    )

    for major, name, _ in for_each_registered_blkdev(prog):
        name_string = escape_ascii_string(name, escape_backslash=True)
        try:
            gendisks = major_to_gendisk[major]
        except KeyError:
            rows.append((major, name_string, CellFormat("(none)", "^")))
        else:
            cell0: Any = major
            cell1: Any = name_string
            # Crash only lists the first gendisk per major, but we list all of
            # them.
            for disk in gendisks:
                ops = disk.fops.value_()
                try:
                    operations_cell = prog.symbol(ops).name
                except LookupError:
                    operations_cell = CellFormat(ops, "x")
                rows.append(
                    (cell0, cell1, CellFormat(disk.value_(), "^x"), operations_cell)
                )
                cell0 = cell1 = ""

    print_table(rows)


def _print_idt(prog: Program, drgn_arg: bool) -> None:
    if prog.platform.arch != Architecture.X86_64:  # type: ignore[union-attr]  # platform can't be None.
        raise ValueError("-d is only supported on x86")

    if drgn_arg:
        sys.stdout.write(
            """\
from drgn.helpers.linux.irq import gate_desc_func


for vec, gate in enumerate(prog["idt_table"]):
    func = gate_desc_func(gate)
    try:
        sym = prog.symbol(func)
    except LookupError:
        pass
"""
        )
        return

    rows: List[Sequence[Any]] = []
    for vec, gate in enumerate(prog["idt_table"]):
        func = gate_desc_func(gate).value_()
        try:
            sym = prog.symbol(func)
        except LookupError:
            func_cell: Any = CellFormat(func, "<x")
        else:
            if sym.address == func:
                func_cell = sym.name
            else:
                func_cell = f"{sym.name}+{func - sym.address}"
        rows.append((CellFormat(f"[{vec}]", ">"), func_cell))
    print_table(rows, sep=" ")


def _print_softirqs(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        sys.stdout.write(
            """\
for vec, softirq_action in enumerate(prog["softirq_vec"]):
    action = softirq_action.action
    try:
        sym = prog.symbol(action)
    except LookupError:
        pass
"""
        )
        return

    rows: List[Sequence[Any]] = [
        (
            CellFormat("SOFTIRQ_VEC", "^"),
            CellFormat("ACTION", "^"),
        )
    ]
    for vec, softirq_action in enumerate(prog["softirq_vec"]):
        action = softirq_action.action.value_()
        try:
            sym = prog.symbol(action)
        except LookupError:
            func_cell = ""
        else:
            func_cell = f"<{sym.name}>"
        rows.append((CellFormat(f"[{vec}]", "^"), CellFormat(action, "^x"), func_cell))
    print_table(rows)


def _print_irq_affinity(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        # Normally we don't bother showing how to format output, but
        # cpumask_to_cpulist() is useful to know.
        sys.stdout.write(
            """\
from drgn.helpers.linux.cpumask import cpumask_to_cpulist
from drgn.helpers.linux.irq import (
    for_each_irq_desc,
    irq_desc_action_names,
    irq_desc_affinity_mask,
)


for irq, irq_desc in for_each_irq_desc():
    if not irq_desc.action:
        continue
    names = irq_desc_action_names(irq_desc)
    affinity = irq_desc_affinity_mask(irq_desc)
    affinity_list = cpumask_to_cpulist(affinity)
"""
        )
        return

    rows: List[Sequence[Any]] = [("IRQ", "NAME", "AFFINITY")]
    for irq, irq_desc in for_each_irq_desc(prog):
        if not irq_desc.action:
            continue
        rows.append(
            (
                irq,
                escape_ascii_string(
                    b",".join(irq_desc_action_names(irq_desc)), escape_backslash=True
                ),
                cpumask_to_cpulist(irq_desc_affinity_mask(irq_desc)),
            )
        )
    print_table(rows)


def _print_irq_stats(prog: Program, drgn_arg: bool, cpuspec: Cpuspec) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        multiple_cpus = code.append_cpuspec_list(cpuspec)

        code.add_from_import(
            "drgn.helpers.linux.irq",
            "for_each_irq_desc",
            "irq_desc_action_names",
            "irq_desc_chip_name",
            "irq_desc_kstat_cpu",
        )
        code.append(
            """\
for irq, irq_desc in for_each_irq_desc():
    if not irq_desc.action:
        continue

"""
        )

        if multiple_cpus:
            code.append("    for cpu in cpus:\n    ")
        code.append(
            """\
    count = irq_desc_kstat_cpu(irq_desc, cpu)
    chip_name = irq_desc_chip_name(irq_desc)
    names = irq_desc_action_names(irq_desc)
"""
        )
        return code.print()

    cpus = cpuspec.cpus(prog)

    header: List[Any] = [""]
    for cpu in cpus:
        header.append(CellFormat(f"CPU{cpu}", ">"))
    rows: List[Sequence[Any]] = [header]

    for irq, irq_desc in for_each_irq_desc(prog):
        if not irq_desc.action:
            continue

        row: List[Any] = [CellFormat(f"{irq}:", ">")]
        # Open-code irq_desc_kstat_cpu() for efficiency. See the comment in
        # that function re: Linux kernel commit 86d2a2f51fba ("genirq: Convert
        # kstat_irqs to a struct") (in v6.10).
        kstat_irqs = irq_desc.kstat_irqs.read_()
        if kstat_irqs:
            try:
                kstat_irqs_cnt = kstat_irqs.cnt
            except AttributeError:
                kstat_irqs_cnt = kstat_irqs[0]
            for cpu in cpus:
                row.append(per_cpu(kstat_irqs_cnt, cpu).value_())
        else:
            row.extend([0] * len(cpus))
        chip_name = irq_desc_chip_name(irq_desc)
        row.append(
            ""
            if chip_name is None
            else escape_ascii_string(chip_name, escape_backslash=True)
        )
        row.append(
            escape_ascii_string(
                b",".join(irq_desc_action_names(irq_desc)), escape_backslash=True
            )
        )
        rows.append(row)

    print_table(rows)


@crash_command(
    description="interrupt requests/descriptors",
    long_description="""
    Show interrupt information. By default, display all interrupt descriptors.
    """,
    arguments=(
        mutually_exclusive_group(
            argument(
                "-u",
                dest="only_used",
                action="store_true",
                help="display only allocated interrupt descriptors",
            ),
            argument(
                "-d",
                dest="idt",
                action="store_true",
                help="display the x86 IDT (interrupt descriptor table)",
            ),
            argument(
                "-b",
                dest="bottom_half",
                action="store_true",
                help="display softirqs (a.k.a. bottom half)",
            ),
            argument(
                "-a",
                dest="affinity",
                action="store_true",
                help="display CPU affinity for in-use interrupts",
            ),
            argument(
                "-s",
                dest="stats",
                action="store_true",
                help="display statistics for in-use interrupts",
            ),
            argument(
                "number",
                type=int,
                nargs="*",
                # Work around https://github.com/python/cpython/issues/72795
                # before Python 3.13.
                default=[],
                help="""
                display the interrupt descriptor for the given IRQ number. May
                be given multiple times
                """,
            ),
        ),
        argument(
            "-c",
            dest="cpu",
            help="""
            when used with -s, restrict the output to one or more CPUs, which
            may be a comma-separated string of CPU numbers or ranges (e.g.,
            '0,3-4')
            """,
        ),
        drgn_argument,
    ),
)
def _crash_cmd_irq(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    *,
    parser: argparse.ArgumentParser,
    **kwargs: Any,
) -> None:
    if args.stats:
        if args.cpu is None:
            args.cpu = "all"
        return _print_irq_stats(prog, args.drgn, parse_cpuspec(args.cpu))
    elif args.cpu is not None:
        parser.error("-c can only be used with -s")

    if args.idt:
        return _print_idt(prog, args.drgn)
    elif args.bottom_half:
        return _print_softirqs(prog, args.drgn)
    elif args.affinity:
        return _print_irq_affinity(prog, args.drgn)

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        indent = "    "
        if args.number:
            code.add_from_import("drgn.helpers.linux.irq", "irq_to_desc")
            if len(args.number) > 1:
                code.append("for irq in (")
                code.append(", ".join([str(number) for number in args.number]))
                code.append("):\n    irq_desc = irq_to_desc(irq)\n")
            else:
                code.append(f"irq_desc = irq_to_desc({args.number[0]})\n")
                indent = ""
        else:
            # Unused descriptors aren't very interesting, so our output with
            # and without -u is the same.
            code.add_from_import("drgn.helpers.linux.irq", "for_each_irq_desc")
            code.append("for irq, irq_desc in for_each_irq_desc():\n")
        code.append(
            f"""\
{indent}action = irq_desc.action.read_()
{indent}while action:
{indent}    name = action.name
{indent}    action = action.next.read_()
"""
        )
        return code.print()

    if args.number:

        def irq_descs(prog: Program) -> Iterator[Tuple[int, Optional[Object]]]:
            nr_irqs = prog["nr_irqs"].value_()
            for number in args.number:
                if number >= nr_irqs:
                    print("irq: invalid IRQ number:", number)
                else:
                    irq_desc = irq_to_desc(prog, number)
                    yield number, irq_desc

    elif args.only_used:
        irq_descs = for_each_irq_desc
    else:

        def irq_descs(prog: Program) -> Iterator[Tuple[int, Optional[Object]]]:
            prev_irq = -1
            for irq, irq_desc in for_each_irq_desc(prog):
                for prev_irq in range(prev_irq + 1, irq):
                    yield prev_irq, None
                prev_irq = irq
                yield irq, irq_desc
            for irq in range(prev_irq + 1, prog["nr_irqs"]):
                yield irq, None

    rows: List[Sequence[Any]] = [
        (
            CellFormat("IRQ", "^"),
            # Crash calls this IRQ_DESC/_DATA, but irq_data is not at the same
            # address as irq_desc since Linux kernel commit 0d0b4c866bcc
            # ("genirq: Introduce struct irq_common_data to host shared irq
            # data") (in v4.2).
            CellFormat("IRQ_DESC", "^"),
            CellFormat("IRQACTION", "^"),
            "NAME",
        )
    ]
    for irq, irq_desc in irq_descs(prog):
        if not irq_desc:
            rows.append(
                (
                    CellFormat(irq, "^"),
                    CellFormat("(unused)", "^"),
                    CellFormat("(unused)", "^"),
                )
            )
            continue

        irq_cell: Any = CellFormat(irq, "^")
        irq_desc_cell: Any = CellFormat(irq_desc.value_(), "^x")
        action = irq_desc.action.read_()
        if action:
            while True:
                action_name = action.name.read_()
                rows.append(
                    (
                        irq_cell,
                        irq_desc_cell,
                        CellFormat(action.value_(), "^x"),
                        (
                            double_quote_ascii_string(action_name.string_())
                            if action_name
                            else ""
                        ),
                    )
                )
                action = action.next.read_()
                if not action:
                    break
                irq_cell = irq_desc_cell = ""
        else:
            rows.append((irq_cell, irq_desc_cell, CellFormat("(unused)", "^")))

    if len(rows) > 1:
        print_table(rows)


@crash_command(
    description="Dump kernel dmesg",
    arguments=(
        mutually_exclusive_group(
            argument(
                "-T",
                dest="timestamps",
                action="store_const",
                const="human",
                default=True,
                help="Dump kernel dmesg in human readable time",
            ),
            argument(
                "-t",
                dest="timestamps",
                action="store_false",
                default=argparse.SUPPRESS,
                help="Dump kernel dmesg without timestamp",
            ),
        ),
        drgn_argument,
    ),
)
def _crash_cmd_log(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.printk", "get_dmesg")
        code.append(
            """\
# Or print_dmesg() if you just want to print it.
dmesg = get_dmesg("""
        )
        if isinstance(args.timestamps, str):
            code.append(f"timestamps={_repr_black(args.timestamps)}")
        elif not args.timestamps:
            code.append("timestamps=False")
        code.append(")\n")
        return code.print()
    print_dmesg(prog, timestamps=args.timestamps)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Block Layer
-----------

The ``drgn.helpers.linux.block`` module provides helpers for working with the
Linux block layer, including disks (``struct gendisk``) and partitions.

Since Linux v5.11, partitions are represented by ``struct block_device``.
Before that, they were represented by ``struct hd_struct``.
"""

from typing import Callable, Iterator, Literal, Optional, Tuple

from drgn import Object, ObjectNotFoundError, Program, cast, container_of
from drgn.helpers.common.format import escape_ascii_string
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.device import MAJOR, MINOR, MKDEV, class_for_each_device
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.sbitmap import sbitmap_for_each_set
from drgn.helpers.linux.xarray import xa_for_each

__all__ = (
    "bdev_partno",
    "blk_mq_rq_from_pdu",
    "blk_mq_rq_to_pdu",
    "blk_rq_bytes",
    "blk_rq_pos",
    "disk_devt",
    "disk_name",
    "for_each_disk",
    "for_each_partition",
    "nr_blockdev_pages",
    "op_is_write",
    "part_devt",
    "part_name",
    "print_disks",
    "print_partitions",
    "req_op",
    "request_queue_busy_iter",
    "rq_data_dir",
)


def disk_devt(disk: Object) -> Object:
    """
    Get a disk's device number.

    :param disk: ``struct gendisk *``
    :return: ``dev_t``
    """
    return Object(disk.prog_, "dev_t", MKDEV(disk.major, disk.first_minor))


def disk_name(disk: Object) -> bytes:
    """
    Get the name of a disk (e.g., ``sda``).

    :param disk: ``struct gendisk *``
    """
    return disk.disk_name.string_()


def _bdev_partno_flags(bdev: Object) -> Object:
    return cast("u8", bdev.__bd_flags.counter)


def _bdev_partno_old(bdev: Object) -> Object:
    return bdev.bd_partno.read_()


def bdev_partno(bdev: Object) -> Object:
    """
    Get the partition number of a block device.

    :param bdev: ``struct block_device *``
    :return: ``u8``
    """
    try:
        impl = bdev.prog_.cache["bdev_partno"]
    except KeyError:
        # Since Linux kernel commit 1116b9fa15c0 ("bdev: infrastructure for
        # flags") (in v6.10), partno is part of the atomic_t __bd_flags member.
        # Before that, it's its own member.
        bdev.prog_.cache["bdev_partno"] = impl = (
            _bdev_partno_flags
            if bdev.prog_.type("struct block_device").has_member("__bd_flags")
            else _bdev_partno_old
        )
    return impl(bdev)


@takes_program_or_default
def for_each_disk(prog: Program) -> Iterator[Object]:
    """
    Iterate over all disks in the system.

    :return: Iterator of ``struct gendisk *`` objects.
    """
    # Before Linux kernel commit 0d02129e76ed ("block: merge struct
    # block_device and struct hd_struct") (in v5.11), partition devices are in
    # struct hd_struct::__dev. After that commit, they are in struct
    # block_device::bd_device. We start by assuming that the kernel has this
    # commit and fall back to the old path if that fails.
    have_bd_device = True
    for device in class_for_each_device(prog["block_class"].address_of_()):
        if have_bd_device:
            try:
                bdev = container_of(device, "struct block_device", "bd_device")
            except LookupError:
                have_bd_device = False
            else:
                if not bdev_partno(bdev):
                    yield bdev.bd_disk
                continue
        part = container_of(device, "struct hd_struct", "__dev")
        if part.partno == 0:
            yield container_of(part, "struct gendisk", "part0")


@takes_program_or_default
def print_disks(prog: Program) -> None:
    """Print all of the disks in the system."""
    for disk in for_each_disk(prog):
        major = disk.major.value_()
        minor = disk.first_minor.value_()
        name = escape_ascii_string(disk_name(disk), escape_backslash=True)
        print(f"{major}:{minor} {name} ({disk.type_.type_name()})0x{disk.value_():x}")


def part_devt(part: Object) -> Object:
    """
    Get a partition's device number.

    :param part: ``struct block_device *`` or ``struct hd_struct *`` depending
        on the kernel version.
    :return: ``dev_t``
    """
    try:
        return part.bd_dev
    except AttributeError:
        return part.__dev.devt


def part_name(part: Object) -> bytes:
    """
    Get the name of a partition (e.g., ``sda1``).

    :param part: ``struct block_device *`` or ``struct hd_struct *`` depending
        on the kernel version.
    """
    try:
        bd_device = part.bd_device
    except AttributeError:
        return part.__dev.kobj.name.string_()
    return bd_device.kobj.name.string_()


@takes_program_or_default
def for_each_partition(prog: Program) -> Iterator[Object]:
    """
    Iterate over all partitions in the system.

    :return: Iterator of ``struct block_device *`` or ``struct hd_struct *``
        objects depending on the kernel version.
    """
    # See the comment in for_each_disk().
    have_bd_device = True
    for device in class_for_each_device(prog["block_class"].address_of_()):
        if have_bd_device:
            try:
                yield container_of(device, "struct block_device", "bd_device")
                continue
            except LookupError:
                have_bd_device = False
        yield container_of(device, "struct hd_struct", "__dev")


@takes_program_or_default
def print_partitions(prog: Program) -> None:
    """Print all of the partitions in the system."""
    for part in for_each_partition(prog):
        devt = part_devt(part).value_()
        name = escape_ascii_string(part_name(part), escape_backslash=True)
        print(
            f"{MAJOR(devt)}:{MINOR(devt)} {name} ({part.type_.type_name()})0x{part.value_():x}"
        )


@takes_program_or_default
def nr_blockdev_pages(prog: Program) -> int:
    """Get the number of memory pages used for block device buffers."""
    return sum(
        inode.i_mapping.nrpages.value_()
        for inode in list_for_each_entry(
            "struct inode",
            prog["blockdev_superblock"].s_inodes.address_of_(),
            "i_sb_list",
        )
    )


def _req_op_impls(
    prog: Program,
) -> Tuple[Callable[[Object], Object], Callable[[Object], bool]]:
    # Since Linux kernel commit ef295ecf090d ("block: better op and flags
    # encoding") (in v4.10), the request operation is the least significant
    # byte of cmd_flags. Before that, it is the most significant 3 bits. We
    # detect this by the presence of enum rq_flag_bits, which was renamed to
    # enum req_flag_bits in that commit.
    #
    # The next commit in that series, 87374179c535 ("block: add a proper block
    # layer data direction encoding"), also changed the operation numbers so
    # that odd numbers are considered writes and even reads. Before that,
    # anything other than REQ_OP_READ (0) was considered a write.
    try:
        prog.type("enum rq_flag_bits")
    except LookupError:
        # The above commit renamed enum req_op to enum req_opf. Commit
        # ff07a02e9e8e ("treewide: Rename enum req_opf into enum req_op") (in
        # v6.0) renamed it back.
        try:
            type = prog.type("enum req_op")
        except LookupError:
            type = prog.type("enum req_opf")

        def req_op(rq: Object) -> Object:
            return cast(type, rq.cmd_flags & 0xFF)

        def op_is_write(op: Object) -> bool:
            return bool(op.value_() & 1)

    else:
        type = prog.type("enum req_op")

        def req_op(rq: Object) -> Object:
            return cast(type, rq.cmd_flags >> 61)

        def op_is_write(op: Object) -> bool:
            return bool(op)

    prog.cache["req_op"] = req_op
    prog.cache["op_is_write"] = op_is_write
    return req_op, op_is_write


def req_op(rq: Object) -> Object:
    """
    Get the operation of a block request.

    >>> req_op(rq)
    (enum req_op)REQ_OP_WRITE

    :param rq: ``struct request *``
    :return: ``enum req_op`` (or ``enum req_opf`` between Linux 4.10 and Linux
        6.0)
    """
    try:
        impl = rq.prog_.cache["req_op"]
    except KeyError:
        impl = _req_op_impls(rq.prog_)[0]
    return impl(rq)


def op_is_write(op: Object) -> bool:
    """
    Return whether a block request operation is a "write"/"data out" operation.

    >>> op_is_write(prog["REQ_OP_READ"])
    False
    >>> op_is_write(prog["REQ_OP_WRITE"])
    True
    >>> op_is_write(prog["REQ_OP_DISCARD"])
    True

    :param op: ``enum req_op`` (or ``enum req_opf`` between Linux 4.10 and
        Linux 6.0)
    """
    try:
        impl = op.prog_.cache["op_is_write"]
    except KeyError:
        impl = _req_op_impls(op.prog_)[1]
    return impl(op)


def rq_data_dir(rq: Object) -> int:
    """
    Return 1 if a block request is a "write"/"data out" operation and 0
    otherwise.

    :param rq: ``struct request *``
    """
    return int(op_is_write(req_op(rq)))


def blk_rq_pos(rq: Object) -> Object:
    """
    Get the current sector of a block request.

    :param rq: ``struct request *``
    :return: ``sector_t``
    """
    return rq.member_("__sector")


def blk_rq_bytes(rq: Object) -> Object:
    """
    Get the number of bytes left in a block request.

    :param rq: ``struct request *``
    :return: ``unsigned int``
    """
    return rq.member_("__data_len")


def blk_mq_rq_to_pdu(rq: Object) -> Object:
    """
    Get the driver command data for a block request.

    :param rq: ``struct request *``
    :return: ``void *``
    """
    return cast("void *", rq + 1)


def blk_mq_rq_from_pdu(pdu: Object) -> Object:
    """
    Get a block request from its driver command data.

    :param pdu: ``void *``
    :return: ``struct request *``
    """
    return cast("struct request *", pdu) - 1


def _queue_for_each_hw_ctx(q: Object) -> Iterator[Object]:
    try:
        queue_hw_ctx = q.queue_hw_ctx
    except AttributeError:
        pass
    else:
        yield from queue_hw_ctx[: q.nr_hw_queues]
        return

    # Between Linux kernel commits 4e5cc99e1e48 ("blk-mq: manage hctx map via
    # xarray") (in v5.18) and d0c98769ee7d ("blk-mq: use array manage hctx map
    # instead of xarray") (in v6.19), hardware contexts are in an xarray.
    hctx_type = q.prog_.type("struct blk_mq_hw_ctx *")
    for _, entry in xa_for_each(q.hctx_table.address_of_()):
        yield cast(hctx_type, entry)


def _blk_mq_tags_for_each(
    tags: Object, rqs: Object, pred: Optional[Callable[[Object, int], bool]]
) -> Iterator[Object]:
    nr_reserved_tags = tags.nr_reserved_tags.value_()
    if nr_reserved_tags:
        for bit in sbitmap_for_each_set(tags.breserved_tags.sb.address_of_()):
            bit += nr_reserved_tags
            rq = rqs[bit].read_()
            if pred is None or pred(rq, bit):
                yield rq

    for bit in sbitmap_for_each_set(tags.bitmap_tags.sb.address_of_()):
        rq = rqs[bit].read_()
        if pred is None or pred(rq, bit):
            yield rq


def request_queue_busy_iter(
    q: Object, tags: Optional[Literal["driver", "sched"]] = None
) -> Iterator[Object]:
    """
    Iterate over all busy requests on a block request queue.

    .. note::

        This does not support the legacy block layer, which was removed in
        Linux 5.0.

    :param q: ``struct request_queue *``
    :param tags: If ``"driver"``, iterate over requests with an allocated
        driver tag. If ``"sched"``, iterate over requests with an allocated
        scheduler tag. Defaults to ``"driver"`` if the I/O scheduler is
        ``none`` and ``"sched"`` otherwise.

        Typically, a scheduler tag is allocated when a request is first
        submitted, then a driver tag is allocated later when the I/O scheduler
        dispatches the request. Both are freed when the request completes. If
        the I/O scheduler is ``none``, then no scheduler tags are allocated.

        Therefore, the default includes all submitted requests regardless of
        the I/O scheduler. (For flush operations, ``"sched"`` will include the
        original request but not the dedicated flush request; see
        :linux:`block/blk-flush.c`.)
    :return: Iterator of ``struct request *`` objects.
    """
    q = q.read_()
    tag_set = q.tag_set.read_()
    if not tag_set:
        # bio-based drivers don't have tags/requests.
        return
    # BLK_MQ_F_TAG_HCTX_SHARED was added in Linux kernel commit 32bc15afed04
    # ("blk-mq: Facilitate a shared sbitmap per tagset") (in v5.10).
    try:
        shared_tags = bool(tag_set.flags & q.prog_["BLK_MQ_F_TAG_HCTX_SHARED"])
    except ObjectNotFoundError:
        shared_tags = False

    if tags != "driver":
        if q.elevator:
            if shared_tags:
                # Since Linux kernel commit e155b0c238b2 ("blk-mq: Use shared
                # tags for shared sbitmap support") (in v5.16), all hctxs with
                # shared tags use the same ->sched_tags.
                try:
                    sched_tags = q.sched_shared_tags.read_()
                except AttributeError:
                    pass
                else:
                    yield from _blk_mq_tags_for_each(
                        sched_tags, sched_tags.static_rqs.read_(), None
                    )
                    return

                # Between that commit and d97e594c5166 ("blk-mq: Use request
                # queue-wide tags for tagset-wide sbitmap") (in v5.14), every
                # hctx with shared tags has its own ->sched_tags but they share
                # the same sbitmaps. Before that, every hctx with shared tags
                # has its own ->sched_tags and sbitmaps. The following works
                # for both cases.
                for hctx in _queue_for_each_hw_ctx(q):
                    sched_tags = hctx.sched_tags.read_()
                    yield from _blk_mq_tags_for_each(
                        sched_tags,
                        sched_tags.static_rqs.read_(),
                        lambda rq, tag: bool(rq.ref.refs.counter),
                    )
                return

            for hctx in _queue_for_each_hw_ctx(q):
                sched_tags = hctx.sched_tags.read_()
                yield from _blk_mq_tags_for_each(
                    sched_tags, sched_tags.static_rqs.read_(), None
                )
            return
        elif tags is not None:
            return

    # Since Linux kernel commit e155b0c238b2 ("blk-mq: Use shared tags for
    # shared sbitmap support") (in v5.16), every hctx with shared tags has its
    # own ->tags but they share the same sbitmaps.
    if shared_tags and hasattr(tag_set, "__bitmap_tags"):
        for hctx in _queue_for_each_hw_ctx(q):
            t = hctx.tags.read_()
            yield from _blk_mq_tags_for_each(
                t,
                t.rqs.read_(),
                lambda rq, tag: bool(rq)
                and rq.tag.value_() == tag
                and rq.mq_hctx == hctx,
            )
        return

    for t in tag_set.tags[: 1 if shared_tags else tag_set.nr_hw_queues]:
        t = t.read_()
        yield from _blk_mq_tags_for_each(
            t,
            t.rqs.read_(),
            lambda rq, tag: bool(rq) and rq.tag.value_() == tag and rq.q == q,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Scheduler-related crash commands."""

import argparse
import sys
import textwrap
from typing import Any, Iterable, List, Sequence

from drgn import Object, Program, offsetof
from drgn.commands import _repr_black, argument, drgn_argument, mutually_exclusive_group
from drgn.commands._crash.common import (
    Cpuspec,
    CrashDrgnCodeBuilder,
    _crash_foreach_subcommand,
    _guess_type,
    _parse_type_name_and_member,
    _prefer_object_lookup,
    _print_task_header,
    _TaskSelector,
    crash_command,
    parse_cpuspec,
    print_task_header,
)
from drgn.helpers.common.format import CellFormat, escape_ascii_string, print_table
from drgn.helpers.linux.pid import for_each_task_in_group
from drgn.helpers.linux.sched import (
    cfs_rq_for_each_entity,
    cpu_rq,
    rq_for_each_rt_task,
    sched_entity_to_task,
    task_group_name,
)
from drgn.helpers.linux.signal import (
    decode_sigaction_flags_value,
    decode_sigset,
    signal_numbers,
    sigpending_for_each,
    sigset_to_hex,
)
from drgn.helpers.linux.wait import waitqueue_for_each_task


def _runq_timestamps(prog: Program, cpuspec: Cpuspec, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)

        code.add_from_import("drgn.helpers.linux.sched", "cpu_rq")
        with code.begin_cpuspec_loop(cpuspec):
            code.append(
                """\
rq = cpu_rq(cpu)
rq_clock = rq.clock
curr = rq.curr.read_()
curr_clock = curr.sched_info.last_arrival
"""
            )
            code.append_task_header(variable="curr", cpu=False)

        return code.print()

    for cpu in cpuspec.cpus(prog):
        rq = cpu_rq(prog, cpu)
        print(f" CPU {cpu}: {rq.clock.value_():016d}")
        curr = rq.curr.read_()
        print(f"        {curr.sched_info.last_arrival.value_():016d}  ", end="")
        _print_task_header(curr, cpu=None)


def _runq_lag(prog: Program, cpuspec: Cpuspec, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.sched", "cpu_rq")
        with code.begin_cpuspec_loop(cpuspec):
            code.append("rq_clock = cpu_rq(cpu).clock\n")
        return code.print()

    timestamps = [(cpu_rq(prog, cpu).clock.value_(), cpu) for cpu in cpuspec.cpus(prog)]
    timestamps.sort(reverse=True)
    if not timestamps:
        return

    cpu_width = max(len(f"CPU {cpu}") for _, cpu in timestamps)
    max_ts = timestamps[0][0]
    for ts, cpu in timestamps:
        cpu_str = f"CPU {cpu}"
        print(f"{cpu_str:>{2 + cpu_width}}: {(max_ts - ts) / 1e9:.2f} secs")


def _elapsed_str(ns: int) -> str:
    ms_total = ns // 1000000
    secs_total, ms = divmod(ms_total, 1000)
    mins_total, secs = divmod(secs_total, 60)
    hours_total, mins = divmod(mins_total, 60)
    days, hours = divmod(hours_total, 24)

    return f"{days} {hours:02}:{mins:02}:{secs:02}.{ms:03}"


def _runq_elapsed(prog: Program, cpuspec: Cpuspec, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)

        code.add_from_import("drgn.helpers.linux.sched", "cpu_rq")
        with code.begin_cpuspec_loop(cpuspec):
            code.append(
                """\
rq = cpu_rq(cpu)
rq_clock = rq.clock
curr = rq.curr.read_()
curr_clock = curr.sched_info.last_arrival
elapsed = rq_clock - curr_clock
"""
            )
            code.append_task_header(variable="curr", cpu=False)

        return code.print()

    entries = []
    max_cpu_width = 0
    max_elapsed_width = 0
    for cpu in cpuspec.cpus(prog):
        rq = cpu_rq(prog, cpu)
        curr = rq.curr.read_()
        elapsed = max(rq.clock.value_() - curr.sched_info.last_arrival.value_(), 0)

        cpu_str = f"CPU {cpu}"
        elapsed_str = _elapsed_str(elapsed)
        entries.append((cpu_str, elapsed_str, curr))

        max_cpu_width = max(max_cpu_width, len(cpu_str))
        max_elapsed_width = max(max_elapsed_width, len(elapsed_str))

    for cpu_str, elapsed_str, curr in entries:
        print(
            f"{cpu_str:>{2 + max_cpu_width}}: [{elapsed_str:>{max_elapsed_width}}]  ",
            end="",
        )
        _print_task_header(curr, cpu=None)


def _print_task_group(cfs_rq: Object, depth: int) -> None:
    tg = cfs_rq.tg.read_()
    indent = " " * (2 + 3 * depth)

    name = task_group_name(tg)
    if name:
        name_str = f"  <{escape_ascii_string(name, escape_backslash=True)}>"
    else:
        name_str = ""

    # throttled only exists if CONFIG_CFS_BANDWIDTH=y.
    if getattr(cfs_rq, "throttled", False):
        throttled_str = " (THROTTLED)"
    else:
        throttled_str = ""

    print(
        f"{indent}TASK_GROUP: {tg.value_():x}  CFS_RQ: {cfs_rq.value_():x}{name_str}{throttled_str}"
    )


def _print_cfs_rq(cfs_rq: Object) -> None:
    found = False
    for se, depth, is_curr, is_task in cfs_rq_for_each_entity(cfs_rq):
        found = True
        if is_task:
            task = sched_entity_to_task(se)
            sys.stdout.write(" " * (5 + 3 * depth))
            _print_task_header(
                task, cpu=None, prio=True, end=" [CURRENT]\n" if is_curr else "\n"
            )
        else:
            _print_task_group(se.my_q.read_(), depth + 1)

    if not found:
        print("     [no tasks queued]")


def _runq_task_groups(prog: Program, cpuspec: Cpuspec, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)

        with code.begin_cpuspec_loop(cpuspec):
            code.append("rq = cpu_rq(cpu)\n")

            code.add_from_import(
                "drgn.helpers.linux.sched",
                "cfs_rq_for_each_entity",
                "cpu_rq",
                "sched_entity_to_task",
                "task_group_name",
            )
            with code.begin_block(
                "for se, depth, is_curr, is_task in cfs_rq_for_each_entity(rq.cfs.address_of_()):\n"
            ):
                with code.begin_block("if is_task:\n"):
                    code.append("task = sched_entity_to_task(se)\n")
                    code.append_task_header(cpu=False, prio=True)
                code.append(
                    """\
else:
    cfs_rq = se.my_q.read_()
    task_group = cfs_rq.tg.read_()
    name = task_group_name(task_group)
    try:
        throttled = cfs_rq.throttled
    except AttributeError:
        # throttled only exists if CONFIG_CFS_BANDWIDTH=y.
        pass
"""
                )

        return code.print()

    root_tg_addr = prog["root_task_group"].address_

    first = True
    for cpu in cpuspec.cpus(prog):
        if first:
            first = False
        else:
            print()

        rq = cpu_rq(prog, cpu)
        curr = rq.curr.read_()

        print(f"CPU {cpu}")
        print("  CURRENT: ", end="")
        _print_task_header(curr, cpu=None)

        cfs_rq = rq.cfs.address_of_()
        print(f"  ROOT_TASK_GROUP: {root_tg_addr:x}" f"  CFS_RQ: {cfs_rq.value_():x}")
        _print_cfs_rq(cfs_rq)


def _print_rq_tasks(tasks: Iterable[Object]) -> None:
    found_task = False
    for task in tasks:
        found_task = True
        sys.stdout.write("     ")
        _print_task_header(task, cpu=None, prio=True)

    if not found_task:
        print("     [no tasks queued]")


def _runq_tasks(prog: Program, cpuspec: Cpuspec, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)

        with code.begin_cpuspec_loop(cpuspec):
            code.append(
                """\
rq = cpu_rq(cpu)

curr = rq.curr.read_()
"""
            )
            code.append_task_header(variable="curr", cpu=False)

            code.add_from_import(
                "drgn.helpers.linux.sched",
                "cfs_rq_for_each_entity",
                "cpu_rq",
                "rq_for_each_rt_task",
                "sched_entity_to_task",
            )
            with code.begin_block(
                """
rt_prio_array = rq.rt.active
for task in rq_for_each_rt_task(rq):
"""
            ):
                code.append_task_header(cpu=False, prio=True)

            with code.begin_block(
                """
cfs_rq = rq.cfs.address_of_()
cfs_rb_root = cfs_rq.tasks_timeline
# Alternatively, you can use rq_for_each_fair_task() if you don't care about
# order or task_group hierarchy.
for se, _, is_curr, is_task in cfs_rq_for_each_entity(cfs_rq):
"""
            ):
                code.append(
                    """\
if is_curr or not is_task:
    continue
task = sched_entity_to_task(se)
"""
                )
                code.append_task_header(cpu=False, prio=True)

        return code.print()

    first = True
    for cpu in cpuspec.cpus(prog):
        if first:
            first = False
        else:
            print()

        rq = cpu_rq(prog, cpu)
        print(f"CPU {cpu} RUNQUEUE {rq.value_():x}")

        print("  CURRENT: ", end="")
        _print_task_header(rq.curr.read_(), cpu=None)

        print(f"  RT PRIO_ARRAY: {rq.rt.active.address_:x}")
        _print_rq_tasks(rq_for_each_rt_task(rq))

        cfs_rq = rq.cfs.address_of_()
        print(f"  CFS RB_ROOT: {cfs_rq.tasks_timeline.address_:x}")
        _print_rq_tasks(
            sched_entity_to_task(se)
            for se, _, is_curr, is_task in cfs_rq_for_each_entity(cfs_rq)
            if not is_curr and is_task
        )


@crash_command(
    description="CPU scheduler run queues",
    long_description="By default, display the tasks on each CPU's runqueue.",
    arguments=(
        mutually_exclusive_group(
            argument(
                "-t",
                dest="timestamps",
                action="store_true",
                help="""
                display the timestamp of each CPU's runqueue and the timestamp
                of the active task on each CPU
                """,
            ),
            argument(
                "-T",
                dest="lag",
                action="store_true",
                help="""
                display the difference between the timestamp of each CPU's
                runqueue relative to the runqueue with the highest timestamp
                """,
            ),
            argument(
                "-m",
                dest="elapsed",
                action="store_true",
                help="""
                display the amount of time that the active task on each CPU has
                been running
                """,
            ),
            argument(
                "-g",
                dest="task_groups",
                action="store_true",
                help="display the task_group hierarchy on each CPU's runqueue",
            ),
        ),
        argument(
            "-c",
            dest="cpu",
            default="all",
            help="restrict the output to one or more CPUs, "
            "which may be a comma-separated string of CPU numbers or ranges "
            "(e.g., '0,3-4')",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_runq(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    cpuspec = parse_cpuspec(args.cpu)

    if args.timestamps:
        _runq_timestamps(prog, cpuspec, args.drgn)
    elif args.lag:
        _runq_lag(prog, cpuspec, args.drgn)
    elif args.elapsed:
        _runq_elapsed(prog, cpuspec, args.drgn)
    elif args.task_groups:
        _runq_task_groups(prog, cpuspec, args.drgn)
    else:
        _runq_tasks(prog, cpuspec, args.drgn)


def _append_sigpending(code: CrashDrgnCodeBuilder, name: str, indent: str = "") -> None:
    code.add_from_import("drgn.helpers.linux.signal", "sigpending_for_each")
    code.append(
        f"""\
{indent}{name}_signals = {name}.signal
{indent}for sigqueue in sigpending_for_each({name}):
{indent}    info = sigqueue.info
{indent}    pending_signo = info.si_signo
"""
    )


def _print_sigpending(pending: Object, indent: str = "") -> None:
    print(indent + "    SIGNAL:", sigset_to_hex(pending.signal))
    rows = []
    first = True
    for sigqueue in sigpending_for_each(pending):
        if first:
            rows.append(
                (
                    indent + "  SIGQUEUE:",
                    CellFormat("SIG", ">"),
                    CellFormat("SIGINFO", "^"),
                )
            )
            first = False
        info = sigqueue.info
        rows.append(("", info.si_signo.value_(), CellFormat(info.address_, "^x")))
    if first:
        print(indent + "  SIGQUEUE: (empty)")
    else:
        print_table(rows)


@_crash_foreach_subcommand(
    arguments=(
        argument(
            "-g",
            dest="thread_group",
            action="store_true",
        ),
        drgn_argument,
    ),
)
def _crash_foreach_sig(task_selector: _TaskSelector, args: argparse.Namespace) -> None:
    prog = task_selector.prog

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            code.append_task_header()
            code.add_from_import("drgn.helpers.linux.signal", "decode_sigaction_flags")
            code.append(
                """\

signal_struct = task.signal
nr_threads = signal_struct.nr_threads

for signo, action in enumerate(task.sighand.action, 1):
    sigaction = action.sa
    handler = sigaction.sa_handler
    mask = sigaction.sa_mask
    flags = sigaction.sa_flags
    decoded_flags = decode_sigaction_flags(sigaction)
"""
            )

            if not args.thread_group:
                code.append(
                    """
blocked = task.blocked

private_pending = task.pending
"""
                )
                _append_sigpending(code, "private_pending")

            code.append(
                """
shared_pending = signal_struct.shared_pending
"""
            )
            _append_sigpending(code, "shared_pending")

            if args.thread_group:
                code.add_from_import("drgn.helpers.linux.pid", "for_each_task_in_group")
                code.append(
                    """
for task in for_each_task_in_group(task, include_self=True):
"""
                )
                code.append_task_header("    ")
                code.append(
                    """
    blocked = task.blocked

    private_pending = task.pending
"""
                )
                _append_sigpending(code, "private_pending", "    ")

        return code.print()

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()
        print_task_header(task)

        signal_struct = task.signal.read_()
        print(
            f"SIGNAL_STRUCT: {signal_struct.value_():x}  NR_THREADS: {signal_struct.nr_threads.value_()}"
        )

        rows: List[Sequence[Any]] = [
            (
                CellFormat("SIG", ">"),
                CellFormat("SIGACTION", "^"),
                CellFormat("HANDLER", "^"),
                CellFormat("MASK", "^"),
                "FLAGS",
            ),
        ]
        for signo, action in enumerate(task.sighand.action, 1):
            sa = action.sa

            handler = sa.sa_handler.value_()
            if handler == 0:
                handler_cell = CellFormat("SIG_DFL", "^")
            elif handler == 1:
                handler_cell = CellFormat("SIG_IGN", "^")
            else:
                handler_cell = CellFormat(handler, ">x")

            flags = sa.sa_flags.value_()
            if flags:
                flags_cell = f"{flags:x} ({decode_sigaction_flags_value(prog, flags)})"
            else:
                flags_cell = "0"

            rows.append(
                (
                    CellFormat(f"[{signo}]", ">"),
                    CellFormat(sa.address_, ">x"),
                    handler_cell,
                    sigset_to_hex(sa.sa_mask),
                    flags_cell,
                )
            )
        print_table(rows)

        # Crash also displays SIGPENDING, which checks whether TIF_SIGPENDING
        # is set on the task. But TIF flags are a pain to get, so we omit it
        # for now.
        if not args.thread_group:
            print("   BLOCKED:", sigset_to_hex(task.blocked))
            print("PRIVATE_PENDING")
            _print_sigpending(task.pending)

        print("SHARED_PENDING")
        _print_sigpending(signal_struct.shared_pending)

        if args.thread_group:
            for thread in for_each_task_in_group(task, include_self=True):
                sys.stdout.write("\n  ")
                print_task_header(thread)
                print("     BLOCKED:", sigset_to_hex(thread.blocked))
                print("  PRIVATE_PENDING")
                _print_sigpending(thread.pending, indent="  ")


@crash_command(
    description="signal handling",
    arguments=(
        mutually_exclusive_group(
            argument(
                "-g",
                dest="thread_group",
                action="store_true",
                help="display pending signals for all threads",
            ),
            argument(
                "-l",
                dest="list",
                action="store_true",
                help="display all known signal numbers and names",
            ),
            argument(
                "-s",
                dest="sigset",
                type="hexadecimal",
                help="translate a hexadecimal signal set into a list of signal names",
            ),
        ),
        argument(
            "tasks",
            metavar="pid|task",
            type="pid_or_task",
            nargs="*",
            help="""
            display signal handlers and pending signals for this task, given as
            either a decimal process ID or a hexadecimal ``task_struct``
            address. May be given multiple times. Defaults to the current
            context
            """,
        ),
        drgn_argument,
    ),
)
def _crash_cmd_sig(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    **kwargs: Any,
) -> None:
    if args.list:
        if args.drgn:
            sys.stdout.write(
                """\
from drgn.helpers.linux.signal import signal_numbers


for number, names in signal_numbers().items():
    ...
"""
            )
            return

        print_table(
            [
                (CellFormat(f"[{number}]", ">"), "/".join(names))
                for number, names in sorted(signal_numbers(prog).items())
            ],
            sep=" ",
        )
        return

    if args.sigset is not None:
        if args.drgn:
            sys.stdout.write(
                f"""\
from drgn.helpers.linux.signal import decode_sigset


decoded = decode_sigset({hex(args.sigset)})
"""
            )
            return

        decoded = decode_sigset(prog, args.sigset)
        decoded = decoded.replace("{", "").replace("}", "").replace(",", " ")
        print(textwrap.fill(decoded, width=80))
        return

    if not args.tasks:
        args.tasks.append(None)
    return _crash_foreach_sig(_TaskSelector(prog, args.tasks), args)


@crash_command(
    description="list tasks on a wait queue",
    usage=r"**waitq** (*symbol* | *address* | *struct.member* *struct_addr*) [**\-\-drgn**]",
    arguments=(
        # argparse can't express symbol | address | struct.member struct_addr,
        # so we use this hack to consume all of the positional arguments (which
        # we parse manually) and define dummy arguments that are only used for
        # the help string.
        argument(
            "posargs",
            nargs="*",
            # Work around https://github.com/python/cpython/issues/72795
            # before Python 3.13.
            default=[],
            help=argparse.SUPPRESS,
        ),
        argument(
            "symbol",
            nargs="?",
            help="symbol name of wait queue",
        ),
        argument(
            "address",
            nargs="?",
            help="hexadecimal address of wait queue",
        ),
        argument(
            "struct.member",
            nargs="?",
            help="""
            name of structure type containing a wait queue member with the
            given name
            """,
        ),
        argument(
            "struct_addr",
            nargs="?",
            help="address of structure containing wait queue member",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_waitq(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    *,
    parser: argparse.ArgumentParser,
    **kwargs: Any,
) -> None:
    is_symbol = False
    if not args.posargs:
        parser.error("symbol, address, or struct is required")
    elif len(args.posargs) == 1:
        try:
            address = int(args.posargs[0], 16)
        except ValueError:
            is_symbol = True
    elif len(args.posargs) == 2:
        type_name, member = _parse_type_name_and_member(args.posargs[0])
        address = int(args.posargs[1], 16)
        try:
            offset_type = _guess_type(prog, type_name)
        except LookupError:
            if not args.drgn:
                raise
            type_name = "struct " + type_name
        else:
            type_name = offset_type.type_name()
    else:
        parser.error(f"unrecognized arguments: {' '.join(args.posargs[2:])}")

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.wait", "waitqueue_for_each_task")

        if len(args.posargs) == 2:
            code.add_from_import("drgn", "Object", "offsetof")
            code.append(
                f"""\
address = {hex(address)}
address += offsetof(prog.type({_repr_black(type_name)}), {_repr_black(member)})
wq = Object(prog, "wait_queue_head_t *", address)
"""
            )
        elif not is_symbol:
            code.add_from_import("drgn", "Object")
            code.append(f'wq = Object(prog, "wait_queue_head_t *", {hex(address)})\n')
        elif _prefer_object_lookup(prog, "wait_queue_head_t", args.posargs[0]):
            code.append(f"wq = prog[{_repr_black(args.posargs[0])}].address_of_()\n")
        else:
            code.add_from_import("drgn", "Object")
            code.append(
                f"""\
address = prog.symbol({_repr_black(args.posargs[0])}).address
wq = Object(prog, "wait_queue_head_t *", address)
"""
            )

        code.append("\nfor task in waitqueue_for_each_task(wq):\n")
        code.append_task_header(indent="    ")
        code.print()
        return

    if len(args.posargs) == 2:
        address += offsetof(offset_type, member)
    elif is_symbol:
        address = prog.symbol(args.posargs[0]).address

    wq = Object(prog, "wait_queue_head_t *", address)
    empty = True
    for task in waitqueue_for_each_task(wq):
        empty = False
        print_task_header(task)
    if empty:
        wq_name = f"{address:x}"
        if is_symbol:
            wq_name = f"{_repr_black(args.posargs[0])} ({wq_name})"
        print(f"wait queue {wq_name} is empty")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Common
------

Linux kernel specializations of :mod:`~drgn.helpers.common` helpers.
"""

import dataclasses
import os
from typing import Any, Dict, Iterable, Iterator, Optional, Tuple, Union

from drgn import (
    NULL,
    Architecture,
    FaultError,
    Object,
    ObjectNotFoundError,
    Program,
    Symbol,
    sizeof,
)
from drgn.helpers.common.format import escape_ascii_string
from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.mm import (
    find_vmap_area,
    for_each_valid_page_range,
    in_direct_map,
)
from drgn.helpers.linux.pid import find_task, for_each_task
from drgn.helpers.linux.sched import idle_task, task_cpu
from drgn.helpers.linux.slab import (
    SlabObjectInfo,
    _find_containing_slab,
    _get_slab_cache_helper,
)

__all__ = ()


@dataclasses.dataclass
class IdentifiedTaskStruct:
    """
    :class:`~drgn.helpers.common.memory.IdentifiedAddress` for an address in a
    ``struct task_struct``.
    """

    address: int

    task: Object
    """``struct task_struct *`` containing the address."""

    def __str__(self) -> str:
        s = (
            f"task: {self.task.pid.value_()}"
            f" ({os.fsdecode(self.task.comm.string_())})"
        )
        task_address = self.task.value_()
        if self.address > task_address:
            s += f" +{hex(self.address - task_address)}"
        return s


@dataclasses.dataclass
class IdentifiedTaskStack:
    """
    :class:`~drgn.helpers.common.memory.IdentifiedAddress` for an address in a
    kernel stack.
    """

    address: int

    task: Object
    """``struct task_struct *`` of the task whose stack contains the address."""

    def __str__(self) -> str:
        return (
            f"task stack: {self.task.pid.value_()}"
            f" ({os.fsdecode(self.task.comm.string_())})"
            f" +{hex(self.address - self.task.stack.value_())}"
        )


@dataclasses.dataclass
class IdentifiedSlabObject:
    """
    :class:`~drgn.helpers.common.memory.IdentifiedAddress` for an address from
    the slab allocator.
    """

    address: int

    slab_object_info: SlabObjectInfo
    """Information about slab object containing the address."""

    def __str__(self) -> str:
        if not self.slab_object_info.address:  # SLOB
            return "unknown slab object"

        cache_name = escape_ascii_string(
            self.slab_object_info.slab_cache.name.string_(), escape_backslash=True
        )
        if self.slab_object_info.allocated:
            maybe_free = ""
        elif self.slab_object_info.allocated is None:
            maybe_free = "corrupted "
        else:
            maybe_free = "free "
        return f"{maybe_free}slab object: {cache_name}+{hex(self.address - self.slab_object_info.address)}"


@dataclasses.dataclass
class IdentifiedVmap:
    """
    :class:`~drgn.helpers.common.memory.IdentifiedAddress` for an address in a
    vmap/vmalloc allocation.
    """

    address: int

    vmap_area: Object
    """``struct vmap_area *`` containing the address."""

    vm_struct: Object
    """``struct vm_struct *`` containing the address."""

    def __str__(self) -> str:
        caller = ""
        caller_value = self.vm_struct.caller.value_()
        try:
            caller_sym = self.vm_struct.prog_.symbol(caller_value)
        except LookupError:
            pass
        else:
            caller = (
                f" caller {caller_sym.name}+{hex(caller_value - caller_sym.address)}"
            )
        return (
            f"vmap: {hex(self.vmap_area.va_start)}-{hex(self.vmap_area.va_end)}{caller}"
        )


@dataclasses.dataclass
class IdentifiedPage:
    """
    :class:`~drgn.helpers.common.memory.IdentifiedAddress` for an address in a
    ``struct page`` object.
    """

    address: int

    page: Object
    """``struct page *`` containing the address."""

    pfn: int
    """Page frame number of the page."""

    def __str__(self) -> str:
        s = f"page: pfn {self.pfn}"
        page_address = self.page.value_()
        if self.address > page_address:
            s += f" +{hex(self.address - page_address)}"
        return s


def _is_task_struct(
    prog: Program, slab_object_info: SlabObjectInfo
) -> Optional[Object]:
    try:
        task_struct_cachep = prog.cache["task_struct_cachep"]
    except KeyError:
        task_struct_cachep = prog["task_struct_cachep"].read_()
        prog.cache["task_struct_cachep"] = task_struct_cachep
    if slab_object_info.slab_cache != task_struct_cachep:
        return None

    task = Object(prog, "struct task_struct *", slab_object_info.address)

    # If the task_struct slab cache is merged, we need to make sure that we got
    # an actual task_struct.
    if task_struct_cachep.refcount.value_() > 1:
        pid = task.pid.value_()
        if pid:
            if find_task(prog, pid) != task:
                return None
        else:
            if idle_task(prog, task_cpu(task)) != task:
                return None

    return task


# Return True if kernel stacks are vmap'ed, the thread_stack_cache Object if
# they are slab allocated, or False if they come from the page allocator.
def _stack_alloc_info(prog: Program) -> Union[bool, Object]:
    try:
        return prog.cache["stack_alloc_mode"]
    except KeyError:
        pass
    if prog.type("struct task_struct").has_member("stack_vm_area"):
        info: Union[bool, Object] = True
    else:
        try:
            info = prog["thread_stack_cache"].read_()
        except ObjectNotFoundError:
            info = False
    prog.cache["stack_alloc_mode"] = info
    return info


# Yield each task_struct that may have an allocated stack.
def _for_each_allocated_task_stack(prog: Program) -> Iterator[Object]:
    yield from for_each_task(prog)
    is_s390x = (
        prog.platform.arch  # type: ignore[union-attr]  # platform can't be None.
        == Architecture.S390X
    )
    for cpu in for_each_online_cpu(prog):
        # s390x between Linux kernel commits ce3dc447493f ("s390: add support
        # for virtually mapped kernel stacks") (in v4.20) and 944c78376a39
        # ("s390: use init_thread_union aka initial stack for the first
        # process") (in v6.4) allocates init_task.stack. In every other case,
        # it is a symbol.
        #
        # Note that on s390x outside of these versions, it's technically
        # incorrect to include init_task here, but it's okay because we don't
        # call this if we already identified a symbol.
        if cpu > 0 or is_s390x:
            yield idle_task(prog, cpu)


def _identify_task_stack(
    prog: Program, addr: int, cache: Optional[Dict[Any, Any]] = None
) -> Iterator[IdentifiedTaskStack]:
    # Kernel stacks are guaranteed to be aligned to the stack size, so we can
    # always compute the stack address that would contain the given address.
    aligned_addr = addr & ~(prog["THREAD_SIZE"].value_() - 1)

    # The cached and uncached cases are separate so that we can avoid creating
    # a large cache and stop early in the uncached case.
    if cache is None:
        for task in _for_each_allocated_task_stack(prog):
            try:
                if task.stack.value_() == aligned_addr:
                    break
            except FaultError:
                continue
        else:
            return
    else:
        try:
            stack_to_task = cache["stack_to_task"]
        except KeyError:
            stack_to_task = {}
            for task in _for_each_allocated_task_stack(prog):
                try:
                    stack_to_task[task.stack.value_()] = task
                except FaultError:
                    pass
            cache["stack_to_task"] = stack_to_task

        try:
            task = stack_to_task[aligned_addr]
        except KeyError:
            return

    yield IdentifiedTaskStack(addr, task)


def _identify_page(
    prog: Program, addr: int, cache: Optional[Dict[Any, Any]] = None
) -> Optional[IdentifiedPage]:
    if cache is None:
        valid_page_ranges: Iterable[Tuple[int, int, Object]] = (
            for_each_valid_page_range(prog)
        )
    else:
        try:
            valid_page_ranges = cache["valid_page_ranges"]
        except KeyError:
            valid_page_ranges = list(for_each_valid_page_range(prog))
            cache["valid_page_ranges"] = valid_page_ranges

    for start_pfn, end_pfn, mem_map in valid_page_ranges:
        start_page = mem_map[start_pfn]
        start_address: int = start_page.address_  # type: ignore[assignment]
        if start_address <= addr < mem_map[end_pfn].address_:  # type: ignore[operator]
            break
    else:
        return None

    sizeof_page = sizeof(start_page)
    index = (addr - start_address) // sizeof_page
    return IdentifiedPage(
        addr,
        Object(prog, mem_map.type_, start_address + index * sizeof_page),
        start_pfn + index,
    )


def _identify_vmap(
    prog: Program, addr: int, check_stack: bool, cache: Optional[Dict[Any, Any]]
) -> Iterator[Union[IdentifiedTaskStack, IdentifiedVmap]]:
    va = find_vmap_area(prog, addr)
    if not va:
        return

    vm = va.vm.read_()
    if not vm:
        return

    # If kernel stacks are vmap'd, check if the address is in a stack.
    if check_stack and _stack_alloc_info(prog) is True:
        yield from _identify_task_stack(prog, addr, cache)

    yield IdentifiedVmap(addr, va, vm)


# Called before identify_address_all() yields an IdentifiedSymbol to check for
# a more specific match. Namely, we want init_task and init_stack to be
# identified as a task or stack, respectively.
def _identify_kernel_symbol(
    prog: Program, addr: int, symbol: Symbol
) -> Iterator[Union[IdentifiedTaskStruct, IdentifiedTaskStack]]:
    task: Object
    init_task_range: Tuple[int, int]
    init_stack_range: Tuple[int, int]
    try:
        task, init_task_range, init_stack_range = prog.cache["init_task_ranges"]
    except KeyError:
        task = prog["init_task"]

        task_address: int = task.address_  # type: ignore
        init_task_range = (task_address, task_address + sizeof(task))

        stack_address = task.stack.value_()
        init_stack_range = (stack_address, stack_address + prog["THREAD_SIZE"].value_())

        task = task.address_of_()
        prog.cache["init_task_ranges"] = (task, init_task_range, init_stack_range)

    if init_task_range[0] <= addr < init_task_range[1]:
        yield IdentifiedTaskStruct(addr, task)
    elif init_stack_range[0] <= addr < init_stack_range[1]:
        yield IdentifiedTaskStack(addr, task)


# Called after identify_address_all() checks if the address is in a symbol. If
# it's not a symbol, we check everything. If it is a symbol, since module
# symbols (and vmlinux symbols on AArch64) are vmapped, we only check vmap.
def _identify_kernel_address(
    prog: Program, addr: int, symbol: bool, cache: Optional[Dict[Any, Any]]
) -> Iterator[
    Union[
        IdentifiedTaskStruct,
        IdentifiedTaskStack,
        IdentifiedSlabObject,
        IdentifiedPage,
        IdentifiedVmap,
    ]
]:
    try:
        direct_map = not symbol and in_direct_map(prog, addr)
    except NotImplementedError:
        # Virtual address translation isn't implemented for this
        # architecture.
        direct_map = None

    if direct_map is not False and "vmemmap" not in prog:
        # Without vmemmap, pages are in the direct mapping.
        identified = _identify_page(prog, addr, cache)
        if identified is not None:
            yield identified
            return

    if direct_map:
        result = _find_containing_slab(prog, addr)
        if result is not None:
            page, slab = result
            try:
                slab_cache = slab.slab_cache.read_()
            except AttributeError:
                # SLOB
                slab_cache = NULL(prog, "struct kmem_cache *")
            slab_info = _get_slab_cache_helper(slab_cache).object_info(page, slab, addr)
            if slab_info:
                if slab_info.allocated:
                    task = _is_task_struct(prog, slab_info)
                    if task is not None:
                        yield IdentifiedTaskStruct(addr, task)
                    else:
                        # If kernel stacks are slab allocated, check if the
                        # address came from the stack slab cache, and if so,
                        # find the stack. Note that this may not find anything
                        # if the allocation was actually from a merged cache.
                        thread_stack_cache = _stack_alloc_info(prog)
                        if (
                            isinstance(thread_stack_cache, Object)
                            and slab_info.slab_cache == thread_stack_cache
                        ):
                            yield from _identify_task_stack(prog, addr, cache)

                yield IdentifiedSlabObject(addr, slab_info)
        elif _stack_alloc_info(prog) is False:
            # If kernel stacks come from the page allocator, check if the
            # address is in a stack.
            yield from _identify_task_stack(prog, addr, cache)
    else:
        if not symbol and "vmemmap" in prog:
            # With vmemmap, pages are outside of the direct mapping.
            identified = _identify_page(prog, addr, cache)
            if identified is not None:
                yield identified
                return
        yield from _identify_vmap(prog, addr, not symbol, cache)
